"""Geodesic solvers implementation."""

from abc import ABC, abstractmethod

import geomstats.backend as gs
from geomstats.numerics.bvp import ScipySolveBVP
from geomstats.numerics.ivp import GSIVPIntegrator
from geomstats.numerics.optimization import ScipyMinimize
from geomstats.numerics.path import (
    UniformlySampledDiscretePath,
    UniformlySampledPathEnergy,
)
from geomstats.vectorization import get_batch_shape


class ExpSolver(ABC):
    """Abstract class for geodesic initial value problem solvers.

    Parameters
    ----------
    solves_ivp : bool
        Informs if solver is able to solve for geodesic at different t.
    """

    def __init__(self, solves_ivp=False):
        self.solves_ivp = solves_ivp

    @abstractmethod
    def exp(self, tangent_vec, base_point):
        """Exponential map.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., dim]
            Tangent vector at the base point.
        base_point : array-like, shape=[..., dim]
            Point on the manifold.

        Returns
        -------
        end_point : array-like, shape=[..., dim]
            Point on the manifold.
        """

    def geodesic_ivp(self, tangent_vec, base_point):
        """Geodesic curve for initial value problem.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., dim]
            Tangent vector at the base point.
        base_point : array-like, shape=[..., dim]
            Point on the manifold.

        Returns
        -------
        path : callable
            Time parametrized geodesic curve. `f(t)`.
        """
        raise NotImplementedError("Can't solve for geodesic at different t.")


class ExpODESolver(ExpSolver):
    """Geodesic initial value problem solver.

    Integrate geodesic equation.

    Parameters
    ----------
    space : Manifold
        Equipped manifold.
    integrator : ODEIVPIntegrator
        Instance of ODEIVP integrator.
    """

    def __init__(self, space, integrator=None):
        self._space = space
        super().__init__()

        if integrator is None:
            integrator = GSIVPIntegrator()

        self._integrator = None
        self.integrator = integrator

    @property
    def integrator(self):
        """An instance of ODEIVPIntegrator."""
        return self._integrator

    @integrator.setter
    def integrator(self, integrator):
        """Set integrator."""
        self.solves_ivp = integrator.tchosen
        self._integrator = integrator

    def _solve(self, tangent_vec, base_point, t_eval=None):
        if base_point.ndim != tangent_vec.ndim:
            base_point = gs.broadcast_to(base_point, tangent_vec.shape)

        state_axis = -(self._space.point_ndim + 1)
        initial_state = gs.stack([base_point, tangent_vec], axis=state_axis)

        force = lambda state, _: self._space.metric.geodesic_equation(state)
        if t_eval is None:
            return self.integrator.integrate(force, initial_state)

        return self.integrator.integrate_t(force, initial_state, t_eval)

    def exp(self, tangent_vec, base_point):
        """Exponential map.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., *space.shape]
            Tangent vector at the base point.
        base_point : array-like, shape=[..., *space.shape]
            Point on the manifold.

        Returns
        -------
        end_point : array-like, shape=[..., *space.shape]
            Point on the manifold.
        """
        result = self._solve(tangent_vec, base_point)
        return self._simplify_exp_result(result)

    def geodesic_ivp(self, tangent_vec, base_point):
        """Geodesic curve for initial value problem.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., *space.shape]
            Tangent vector at the base point.
        base_point : array-like, shape=[..., *space.shape]
            Point on the manifold.

        Returns
        -------
        path : callable
            Time parametrized geodesic curve. `f(t)`.
        """
        if not self.solves_ivp:
            raise NotImplementedError(
                "Can't solve for geodesic at different t with this integrator."
            )

        base_point = gs.broadcast_to(base_point, tangent_vec.shape)

        def path(t):
            """Time parametrized geodesic curve.

            Parameters
            ----------
            t : float or array-like, shape=[n_times,]

            Returns
            -------
            geodesic_points : array-like, shape=[..., n_times, *space.shape]
                Geodesic points evaluated at t.
            """
            if not gs.is_array(t):
                t = gs.array([t])

            if gs.ndim(t) == 0:
                t = gs.expand_dims(t, axis=0)

            result = self._solve(tangent_vec, base_point, t_eval=t)
            return self._simplify_result_t(result)

        return path

    def _simplify_exp_result(self, result):
        y = result.get_last_y()
        point_ndim_slc = tuple([slice(None)] * self._space.point_ndim)
        return y[(..., 0) + point_ndim_slc]

    def _simplify_result_t(self, result):
        # assumes several t
        y = result.y
        point_ndim_slc = tuple([slice(None)] * self._space.point_ndim)
        return y[(..., slice(None), 0) + point_ndim_slc]


class LogSolver(ABC):
    """Abstract class for geodesic boundary value problem solvers.

    Parameters
    ----------
    solves_bvp : bool
        Informs if solver is able to solve for geodesic at different t.
    """

    def __init__(self, solves_bvp=False):
        self.solves_bvp = solves_bvp

    @abstractmethod
    def log(self, space, point, base_point):
        """Logarithm map.

        Parameters
        ----------
        end_point : array-like, shape=[..., dim]
            Point on the manifold.
        base_point : array-like, shape=[..., dim]
            Point on the manifold.

        Returns
        -------
        tangent_vec : array-like, shape=[..., dim]
            Tangent vector at the base point.
        """

    def geodesic_bvp(self, point, base_point):
        """Geodesic curve for boundary value problem.

        Parameters
        ----------
        end_point : array-like, shape=[..., dim]
            Point on the manifold.
        base_point : array-like, shape=[..., dim]
            Point on the manifold.

        Returns
        -------
        path : callable
            Time parametrized geodesic curve. `f(t)`.
        """
        raise NotImplementedError("Can't solve for geodesic at different t.")


class _LogBatchMixins:
    """Provides method to compute log for multiples point."""

    @abstractmethod
    def _log_single(self, point, base_point):
        """Logarithm map.

        Parameters
        ----------
        end_point : array-like, shape=[dim]
            Point on the manifold.
        base_point : array-like, shape=[dim]
            Point on the manifold.

        Returns
        -------
        tangent_vec : array-like, shape=[dim]
            Tangent vector at the base point.
        """

    def log(self, point, base_point):
        """Logarithm map.

        Parameters
        ----------
        space : Manifold
            Equipped manifold.
        end_point : array-like, shape=[..., dim]
            Point on the manifold.
        base_point : array-like, shape=[..., dim]
            Point on the manifold.

        Returns
        -------
        tangent_vec : array-like, shape=[..., dim]
            Tangent vector at the base point.
        """
        # assumes inability to properly vectorize
        if point.ndim != base_point.ndim:
            point, base_point = gs.broadcast_arrays(point, base_point)

        is_batch = point.ndim > self._space.point_ndim
        if not is_batch:
            return self._log_single(point, base_point)

        return gs.stack(
            [
                self._log_single(point_, base_point_)
                for point_, base_point_ in zip(point, base_point)
            ]
        )


class LogShootingSolver:
    """Geodesic boundary value problem solver using shooting.

    Parameters
    ----------
    space : Manifold
        Equipped manifold.
    optimizer : ScipyMinimize
        Instance of ScipyMinimize.
    initialization : callable
        Function to provide initial solution. `f(point, base_point)`.
        Defaults to linear initialization.
    flatten : bool
        If True, the optimization problem is solved together for all the batch points.
    """

    def __new__(cls, space, optimizer=None, initialization=None, flatten=True):
        """Instantiate a log shooting solver."""
        if flatten:
            return _LogShootingSolverFlatten(
                space=space,
                optimizer=optimizer,
                initialization=initialization,
            )

        return _LogShootingSolverUnflatten(
            space=space,
            optimizer=optimizer,
            initialization=initialization,
        )


class _LogShootingSolver(LogSolver, ABC):
    """Geodesic boundary value problem solver using shooting.

    Parameters
    ----------
    space : Manifold
        Equipped manifold.
    optimizer : ScipyMinimize
        Instance of ScipyMinimize.
    initialization : callable
        Function to provide initial solution. `f(point, base_point)`.
        Defaults to linear initialization.
    """

    def __init__(self, space, optimizer=None, initialization=None):
        super().__init__(solves_bvp=False)
        self._space = space

        if optimizer is None:
            optimizer = ScipyMinimize(autodiff_jac=True)

        if initialization is None:
            initialization = self._default_initialization

        self.optimizer = optimizer
        self.initialization = initialization

    def _default_initialization(self, point, base_point):
        """Linear initialization.

        Parameters
        ----------
        end_point : array-like, shape=[..., *space.shape]
            Point on the manifold.
        base_point : array-like, shape=[..., *space.shape]
            Point on the manifold.
        """
        return gs.flatten(point - base_point)

    def _objective(self, flat_tangent_vec, point, base_point, batch_shape):
        """Objective function.

        Minimizes Euclidean distance to shooted point.

        Parameters
        ----------
        flat_tangent_vec : array-like, shape=[prod(batch_shape)*prod(space.shape)]
            Flattened tangent vector.
        end_point : array-like, shape=[..., *space.shape]
            Point on the manifold.
        base_point : array-like, shape=[..., *space.shape]
            Point on the manifold.
        batch_shape : tuple
            Batch shape.
        """
        tangent_vec = gs.reshape(flat_tangent_vec, batch_shape + self._space.shape)
        delta = self._space.metric.exp(tangent_vec, base_point) - point
        return gs.sum(delta**2)


class _LogShootingSolverFlatten(_LogShootingSolver):
    """Geodesic boundary value problem solver using shooting.

    Parameters
    ----------
    space : Manifold
        Equipped manifold.
    optimizer : ScipyMinimize
        Instance of ScipyMinimize.
    initialization : callable
        Function to provide initial solution. `f(point, base_point)`.
        Defaults to linear initialization.

    Notes
    -----
    Differs from `_LogShootingSolverUnflatten` as it always solves one
    optimization problem, even in batch mode.
    """

    def log(self, point, base_point):
        """Logarithm map.

        Parameters
        ----------
        end_point : array-like, shape=[..., *space.shape]
            Point on the manifold.
        base_point : array-like, shape=[..., *space.shape]
            Point on the manifold.

        Returns
        -------
        tangent_vec : array-like, shape=[..., *space.shape]
            Tangent vector at the base point.
        """
        batch_shape = get_batch_shape(self._space.point_ndim, point, base_point)

        objective = lambda velocity: self._objective(
            velocity, point, base_point, batch_shape
        )
        init_tangent_vec = self.initialization(point, base_point)

        res = self.optimizer.minimize(objective, init_tangent_vec)

        return gs.reshape(res.x, batch_shape + self._space.shape)


class _LogShootingSolverUnflatten(_LogBatchMixins, _LogShootingSolver):
    """Geodesic boundary value problem solver using shooting.

    Parameters
    ----------
    space : Manifold
        Equipped manifold.
    optimizer : ScipyMinimize
        Instance of ScipyMinimize.
    initialization : callable
        Function to provide initial solution. `f(point, base_point)`.
        Defaults to linear initialization.

    Notes
    -----
    Differs from `_LogShootingSolverFlatten` as it solves one
    optimization problem for each combination of point and base point.
    """

    def _log_single(self, point, base_point):
        """Logarithm map.

        Parameters
        ----------
        end_point : array-like, shape=[*space.shape]
            Point on the manifold.
        base_point : array-like, shape=[*space.shape]
            Point on the manifold.

        Returns
        -------
        tangent_vec : array-like, shape=[*space.shape]
            Tangent vector at the base point.
        """
        objective = lambda velocity: self._objective(
            velocity, point, base_point, batch_shape=()
        )
        init_tangent_vec = self.initialization(point, base_point)

        res = self.optimizer.minimize(objective, init_tangent_vec)

        if self._space.point_ndim > 1:
            return gs.reshape(res.x, self._space.shape)

        return res.x


class LogODESolver(_LogBatchMixins, LogSolver):
    """Geodesic boundary value problem using an ODE solver.

    Parameters
    ----------
    space : Manifold
        Equipped manifold.
    n_nodes : Number of mesh nodes.
    integrator : ScipySolveBVP
        Instance of ScipySolveBVP.
    initialization : callable
        Function to provide initial solution. `f( point, base_point)`.
        Defaults to linear initialization.
    """

    def __init__(
        self, space, n_nodes=10, integrator=None, initialization=None, use_jac=True
    ):
        self._space = space
        super().__init__(solves_bvp=True)

        if integrator is None:
            integrator = ScipySolveBVP()

        if initialization is None:
            initialization = self._default_initialization

        self.n_nodes = n_nodes
        self.integrator = integrator
        self.initialization = initialization
        self.use_jac = use_jac

        self.grid = self._create_grid()

    def _create_grid(self):
        return gs.linspace(0.0, 1.0, num=self.n_nodes)

    def _default_initialization(self, point, base_point):
        if point.ndim == 1:
            point_0, point_1 = base_point, point
        else:
            point_0 = gs.flatten(base_point)
            point_1 = gs.flatten(point)

        pos_init = gs.transpose(gs.linspace(point_0, point_1, self.n_nodes))

        vel_init = self.n_nodes * (pos_init[:, 1:] - pos_init[:, :-1])
        vel_init = gs.hstack([vel_init, vel_init[:, [-2]]])

        return gs.vstack([pos_init, vel_init])

    def _boundary_condition(self, state_0, state_1, point_0, point_1):
        pos_0 = state_0[: point_0.shape[0]]
        pos_1 = state_1[: point_1.shape[0]]
        return gs.hstack((pos_0 - point_0, pos_1 - point_1))

    def _bvp(self, raveled_state):
        """Boundary value problem.

        Parameters
        ----------
        raveled_state : array-like, shape=[2*dim, n_nodes]
            Vector of state variables (position and speed).

        Returns
        -------
        sol : array-like, shape=[2*dim, n_nodes]
        """
        state = gs.moveaxis(
            gs.reshape(raveled_state, (2,) + self._space.shape + (-1,)), -1, 0
        )
        new_state = self._space.metric.geodesic_equation(state)
        return gs.moveaxis(gs.reshape(new_state, (-1, raveled_state.shape[0])), -2, -1)

    def _jacobian(self, _, raveled_state):
        """Jacobian of boundary value problem.

        Parameters
        ----------
        _ : float
            Unused.
        raveled_state : array-like, shape=[2*dim, n_nodes]
            Vector of state variables (position and speed).

        Returns
        -------
        jac : array-like, shape=[dim, dim, n_nodes]
        """
        dim = self._space.dim
        n_nodes = raveled_state.shape[-1]
        position, velocity = gs.transpose(raveled_state[:dim]), raveled_state[dim:]

        dgamma = self._space.metric.jacobian_christoffels(position)

        df_dposition = -gs.einsum(
            "j...,...ijkl,k...->il...", velocity, dgamma, velocity
        )

        gamma = self._space.metric.christoffels(position)
        df_dvelocity = -2 * gs.einsum("...ijk,k...->ij...", gamma, velocity)

        jac_nw = gs.zeros((dim, dim, raveled_state.shape[1]))
        jac_ne = gs.squeeze(gs.transpose(gs.tile(gs.eye(dim), (n_nodes, 1, 1))))
        jac_sw = df_dposition
        jac_se = df_dvelocity
        jac = gs.concatenate(
            (
                gs.concatenate((jac_nw, jac_ne), axis=1),
                gs.concatenate((jac_sw, jac_se), axis=1),
            ),
            axis=0,
        )

        return jac

    def _solve(self, point, base_point):
        bvp = lambda t, state: self._bvp(state)
        bc = lambda state_0, state_1: self._boundary_condition(
            state_0, state_1, gs.flatten(base_point), gs.flatten(point)
        )

        jacobian = None
        if self.use_jac:
            jacobian = lambda t, state: self._jacobian(t, state)

        y = self.initialization(point, base_point)

        return self.integrator.integrate(bvp, bc, self.grid, y, fun_jac=jacobian)

    def _log_single(self, point, base_point):
        res = self._solve(point, base_point)
        return self._simplify_log_result(res)

    def geodesic_bvp(self, point, base_point):
        """Geodesic curve for boundary value problem.

        Parameters
        ----------
        end_point : array-like, shape=[..., dim]
            Point on the manifold.
        base_point : array-like, shape=[..., dim]
            Point on the manifold.

        Returns
        -------
        path : callable
            Time parametrized geodesic curve. `f(t)`. 0 <= t <= 1.
        """
        if point.ndim != base_point.ndim:
            point, base_point = gs.broadcast_arrays(point, base_point)

        is_batch = point.ndim > self._space.point_ndim
        if not is_batch:
            result = self._solve(point, base_point)
        else:
            results = [
                self._solve(point_, base_point_)
                for point_, base_point_ in zip(point, base_point)
            ]

        def path(t):
            """Time parametrized geodesic curve.

            Parameters
            ----------
            t : float or array-like, shape=[n_times,]

            Returns
            -------
            geodesic_points : array-like, shape=[..., n_times, dim]
                Geodesic points evaluated at t.
            """
            if not gs.is_array(t):
                t = gs.array([t])

            if gs.ndim(t) == 0:
                t = gs.expand_dims(t, axis=0)

            if not is_batch:
                return self._simplify_result_t(result.sol(t))

            return gs.array(
                [self._simplify_result_t(result.sol(t)) for result in results]
            )

        return path

    def _simplify_log_result(self, result):
        return gs.reshape(result.y[..., 0], (2,) + self._space.shape)[1]

    def _simplify_result_t(self, result):
        return gs.moveaxis(
            gs.reshape(result[: result.shape[0] // 2, :], self._space.shape + (-1,)),
            -1,
            0,
        )


class _DiscreteGeodesicBVPBatchMixins:
    @abstractmethod
    def _discrete_geodesic_bvp_single(self, point, base_point):
        """Solve boundary value problem (BVP).

        Given an initial point and an end point, solve the geodesic equation
        via minimizing the Riemannian path energy.

        Parameters
        ----------
        point : array-like, shape=[*point_shape]
            Point on the manifold.
        base_point : array-like, shape=[*point_shape]
            Point on the manifold.

        Returns
        -------
        discr_geod_path : array-like, shape=[n_nodes, *point_shape]
            Discrete geodesic.
        """

    def discrete_geodesic_bvp(self, point, base_point):
        """Solve boundary value problem (BVP).

        Given an initial point and an end point, solve the geodesic equation
        via minimizing the Riemannian path energy.

        Parameters
        ----------
        point : array-like, shape=[..., *point_shape]
            Point on the manifold.
        base_point : array-like, shape=[..., *point_shape]
            Point on the manifold.

        Returns
        -------
        discr_geod_path : array-like, shape=[..., n_nodes, *point_shape]
            Discrete geodesic.
        """
        if point.ndim != base_point.ndim:
            point, base_point = gs.broadcast_arrays(point, base_point)

        is_batch = point.ndim > self._space.point_ndim
        if not is_batch:
            return self._discrete_geodesic_bvp_single(point, base_point)

        return gs.stack(
            [
                self._discrete_geodesic_bvp_single(point_, base_point_)
                for point_, base_point_ in zip(point, base_point)
            ]
        )


class PathBasedLogSolver(LogSolver, ABC):
    """A geodesic BVP solver based on finding a discrete geodesic path.

    Parameters
    ----------
    space : Manifold
        Equipped manifold.
    """

    def __init__(self, space):
        self._space = space
        super().__init__(solves_bvp=True)

    def path_n_nodes(self, path):
        """Get number of nodes of a path.

        Parameters
        ----------
        discr_path : array-like, shape=[..., n_nodes, *point_shape]
            Discrete path.

        Returns
        -------
        n_nodes : int
            Number of path nodes.
        """
        return path.shape[-self._space.point_ndim - 1]

    @abstractmethod
    def discrete_geodesic_bvp(self, point, base_point):
        """Solve boundary value problem (BVP).

        Given an initial point and an end point, solve the geodesic equation
        via minimizing the Riemannian path energy.

        Parameters
        ----------
        point : array-like, shape=[..., *point_shape]
            Point on the manifold.
        base_point : array-like, shape=[..., *point_shape]
            Point on the manifold.

        Returns
        -------
        discr_geod_path : array-like, shape=[..., n_nodes, *point_shape]
            Discrete geodesic.
        """

    def log(self, point, base_point):
        """Logarithm map.

        Parameters
        ----------
        end_point : array-like, shape=[..., *point_shape]
            Point on the manifold.
        base_point : array-like, shape=[..., *point_shape]
            Point on the manifold.

        Returns
        -------
        tangent_vec : array-like, shape=[..., *point_shape]
            Tangent vector at the base point.
        """
        discr_geod_path = self.discrete_geodesic_bvp(point, base_point)
        point_ndim_slc = (slice(None),) * self._space.point_ndim
        return (self.path_n_nodes(discr_geod_path) - 1) * (
            discr_geod_path[(..., 1) + point_ndim_slc]
            - discr_geod_path[(..., 0) + point_ndim_slc]
        )

    def geodesic_bvp(self, point, base_point):
        """Geodesic curve for boundary value problem.

        Parameters
        ----------
        end_point : array-like, shape=[..., *point_shape]
            Point on the manifold.
        base_point : array-like, shape=[..., *point_shape]
            Point on the manifold.

        Returns
        -------
        path : callable
            Time parametrized geodesic curve. `f(t)`.
        """
        discr_geod_path = self.discrete_geodesic_bvp(point, base_point)
        return UniformlySampledDiscretePath(
            discr_geod_path, point_ndim=self._space.point_ndim
        )


class PathStraightening(_DiscreteGeodesicBVPBatchMixins, PathBasedLogSolver):
    """Geodesic boundary value problem with path-straightening.

    Parameters
    ----------
    space : Manifold
        Equipped manifold.
    path_energy : callable
        Method to compute Riemannian path energy.
    n_nodes : int
        Number of path discretization points.
    optimizer : ScipyMinimize
        An optimizer to solve path energy minimization problem.
    initialization : callable
        A method to get initial guess for optimization.
    symmetric : bool
        If to use a symmetrized version of the energy.

    References
    ----------
    .. [HSKCB2022] "Elastic shape analysis of surfaces with second-order
        Sobolev metrics: a comprehensive numerical framework".
        arXiv:2204.04238 [cs.CV], 25 Sep 2022
    """

    def __init__(
        self,
        space,
        path_energy=None,
        n_nodes=100,
        optimizer=None,
        initialization=None,
        symmetric=False,
    ):
        super().__init__(space)
        if optimizer is None:
            optimizer = ScipyMinimize(
                method="L-BFGS-B",
                autodiff_jac=True,
                options={"disp": False},
            )

        if path_energy is None:
            path_energy = UniformlySampledPathEnergy(space)

        if initialization is None:
            initialization = self._default_initialization

        self.n_nodes = n_nodes
        self.optimizer = optimizer
        self.path_energy = path_energy
        self.initialization = initialization
        self.symmetric = symmetric

    def _default_initialization(self, point, base_point):
        """Linear initialization.

        Parameters
        ----------
        point : array-like, shape=[..., *point_shape]
            Point on the manifold.
        base_point : array-like, shape=[..., *point_shape]
            Point on the manifold.

        Returns
        -------
        path : array-like, shape=[..., n_nodes, *point_shape]
            Linear path between base point and point.
        """
        times = gs.linspace(0.0, 1.0, self.n_nodes)
        linear_deformation = point - base_point
        return base_point + gs.einsum("t,...->t...", times, linear_deformation)

    def _make_objective(self, point, base_point):
        """Create objective function.

        Parameters
        ----------
        point : array-like, shape=[*point_shape]
            Point on the manifold.
        base_point : array-like, shape=[*point_shape]
            Point on the manifold.

        Returns
        -------
        objective : callable
            Objective function.
        """

        def objective(midpoints):
            """Compute path energy of paths going through a midpoint.

            Parameters
            ----------
            midpoint : array-like, shape=[(self.n_nodes-2) * math.prod(*point_shape)]
                Midpoints of the path.

            Returns
            -------
            _ : float
                Energy of the path.
            """
            midpoints = gs.reshape(midpoints, (self.n_nodes - 2,) + self._space.shape)
            path = gs.concatenate(
                [
                    base_point,
                    midpoints,
                    point,
                ],
            )
            energy = self.path_energy(path)
            if self.symmetric:
                reversed_path = gs.flip(path, axis=0)
                energy = (energy + self.path_energy(reversed_path)) / 2

            return energy

        return objective

    def _discrete_geodesic_bvp_single(self, point, base_point):
        """Solve boundary value problem (BVP).

        Given an initial point and an end point, solve the geodesic equation
        via minimizing the Riemannian path energy.

        Parameters
        ----------
        point : array-like, shape=[*point_shape]
            Point on the manifold.
        base_point : array-like, shape=[*point_shape]
            Point on the manifold.

        Returns
        -------
        discr_geod_path : array-like, shape=[n_nodes, *point_shape]
            Discrete geodesic.
        """
        init_path = self.initialization(point, base_point)
        init_midpoints = init_path[1:-1]

        base_point = gs.expand_dims(base_point, axis=0)
        point = gs.expand_dims(point, axis=0)

        init_midpoints = gs.reshape(init_midpoints, (-1,))

        objective = self._make_objective(point, base_point)
        sol = self.optimizer.minimize(objective, init_midpoints)

        solution_midpoints = gs.reshape(
            gs.array(sol.x), (self.n_nodes - 2,) + self._space.shape
        )

        return gs.concatenate(
            [
                base_point,
                solution_midpoints,
                point,
            ],
            axis=0,
        )


class MultiresPathStraightening(_DiscreteGeodesicBVPBatchMixins, PathBasedLogSolver):
    """Geodesic boundary value problem with multiresolution path straightening.

    Parameters
    ----------
    space : Manifold
        Equipped manifold.
    path_energy : callable
        Method to compute Riemannian path energy.
    n_nodes : list
        Number of path discretization points at each resolution.
    n_levels : int
        Number of resolutions to use. Sets number of nodes following a sequence.
        Ignored if ``n_nodes`` is not None.
    optimizer : ScipyMinimize
        An optimizer to solve path energy minimization problem.
    initialization : callable
        A method to get initial guess for optimization.
    symmetric : bool
        If to use a symmetrized version of the energy.
    early_stop : bool
        If to stop at a resolution if previous resolution was solved
        with one iteration.
    """

    def __init__(
        self,
        space,
        path_energy=None,
        n_nodes=None,
        n_levels=3,
        optimizer=None,
        initialization=None,
        symmetric=False,
        early_stop=False,
    ):
        super().__init__(space)
        if optimizer is None:
            optimizer = ScipyMinimize(
                method="L-BFGS-B",
                autodiff_jac=True,
                options={"disp": False},
            )

        if n_nodes is None:
            n_nodes = self._create_levels_sequence(3, n_levels)

        if path_energy is None:
            path_energy = UniformlySampledPathEnergy(space)

        if initialization is None:
            initialization = self._default_initialization

        self.n_nodes = n_nodes
        self.optimizer = optimizer
        self.path_energy = path_energy
        self.initialization = initialization
        self.symmetric = symmetric
        self.early_stop = early_stop

    @staticmethod
    def _create_levels_sequence(init_val, n_levels):
        seq = [init_val]
        for _ in range(n_levels - 1):
            seq.append(seq[-1] * 2 - 1)

        return seq

    def _default_initialization(self, point, base_point):
        """Linear initialization.

        Parameters
        ----------
        point : array-like, shape=[..., *point_shape]
            Point on the manifold.
        base_point : array-like, shape=[..., *point_shape]
            Point on the manifold.

        Returns
        -------
        path : array-like, shape=[..., n_nodes, *point_shape]
            Linear path between base point and point.
        """
        times = gs.linspace(0.0, 1.0, self.n_nodes[0])
        linear_deformation = point - base_point
        return base_point + gs.einsum("t,...->t...", times, linear_deformation)

    def _initialization_from_previous(self, path, n_nodes):
        """Initialize next resolution using previous resolution solution.

        Parameters
        ----------
        path : array-like, shape=[..., n_nodes_previous, *point_shape]
            Discrete path.

        Returns
        -------
        interp_path : array-like, shape=[..., n_nodes, *point_shape]
            Discrete path.
        """
        discr_geod_path = UniformlySampledDiscretePath(
            path, point_ndim=self._space.point_ndim
        )
        times = gs.linspace(0.0, 1.0, n_nodes)
        return discr_geod_path(times)

    def _make_objective(self, point, base_point, n_nodes):
        """Create objective function.

        Parameters
        ----------
        point : array-like, shape=[*point_shape]
            Point on the manifold.
        base_point : array-like, shape=[*point_shape]
            Point on the manifold.

        Returns
        -------
        objective : callable
            Objective function.
        """

        def objective(midpoints):
            """Compute path energy of paths going through a midpoint.

            Parameters
            ----------
            midpoint : array-like, shape=[(n_nodes-2) * math.prod(*point_shape)]
                Midpoints of the path.

            Returns
            -------
            _ : float
                Energy of the path.
            """
            midpoints = gs.reshape(midpoints, (n_nodes - 2,) + self._space.shape)
            path = gs.concatenate(
                [
                    base_point,
                    midpoints,
                    point,
                ],
            )
            energy = self.path_energy(path)
            if self.symmetric:
                reversed_path = gs.flip(path, axis=0)
                energy = (energy + self.path_energy(reversed_path)) / 2

            return energy

        return objective

    def _discrete_geodesic_bvp_single(self, point, base_point):
        """Solve boundary value problem (BVP).

        Given an initial point and an end point, solve the geodesic equation
        via minimizing the Riemannian path energy.

        Parameters
        ----------
        point : array-like, shape=[*point_shape]
            Point on the manifold.
        base_point : array-like, shape=[*point_shape]
            Point on the manifold.

        Returns
        -------
        discr_geod_path : array-like, shape=[n_nodes, *point_shape]
            Discrete geodesic.
        """
        init_path = self.initialization(point, base_point)
        init_midpoints = init_path[1:-1]

        base_point = gs.expand_dims(base_point, axis=0)
        point = gs.expand_dims(point, axis=0)

        discrete_path = None
        for n_nodes in self.n_nodes:
            objective = self._make_objective(point, base_point, n_nodes)

            if discrete_path is None:
                init_midpoints = self.initialization(point, base_point)[1:-1]

            else:
                init_midpoints = self._initialization_from_previous(
                    discrete_path, n_nodes
                )[1:-1]

            init_midpoints = gs.reshape(init_midpoints, (-1,))
            res = self.optimizer.minimize(objective, init_midpoints)

            path_midpoints = gs.reshape(
                gs.array(res.x), (n_nodes - 2,) + self._space.shape
            )
            discrete_path = gs.concatenate(
                [
                    base_point,
                    path_midpoints,
                    point,
                ],
                axis=0,
            )

            if self.early_stop and res.nit == 1:
                break

        return discrete_path



"""Complex manifold module.

In other words, a topological space that locally resembles
Hermitian space near each point.

Lead author: Yann Cabanes.
"""

import abc
import inspect

import geomstats.backend as gs


class ComplexManifold(abc.ABC):
    r"""Class for complex manifolds.

    Parameters
    ----------
    dim : int
        Dimension of the manifold.
    shape : tuple of int
        Shape of one element of the manifold.
        Optional, default : None.
    intrinsic : bool
        Coordinates type.
    equip : bool
        If True, equip space with default metric.
    """

    def __init__(
        self,
        dim,
        shape,
        intrinsic=True,
        equip=True,
    ):
        self.dim = dim
        self.shape = shape
        self.intrinsic = intrinsic

        self.point_ndim = len(self.shape)

        if equip:
            self.equip_with_metric()

    def equip_with_metric(self, Metric=None, **metric_kwargs):
        """Equip manifold with Metric.

        Parameters
        ----------
        Metric : Connection object or instance or ScalarProductMetric instance
            If None, default metric will be used.
        """
        if Metric is None:
            out = self.default_metric()
            if isinstance(out, tuple):
                Metric, kwargs = out
                kwargs.update(metric_kwargs)
                metric_kwargs = kwargs
            else:
                Metric = out

        if inspect.isclass(Metric):
            self.metric = Metric(self, **metric_kwargs)
        else:
            if self.metric._space is not self:
                raise ValueError(
                    "Cannot equip space with metric instantiated with another space."
                )

            self.metric = Metric

        return self

    @abc.abstractmethod
    def belongs(self, point, atol=gs.atol):
        """Evaluate if a point belongs to the manifold.

        Parameters
        ----------
        point : array-like, shape=[..., *point_shape]
            Point to evaluate.
        atol : float
            Absolute tolerance.
            Optional, default: backend atol.

        Returns
        -------
        belongs : array-like, shape=[...,]
            Boolean evaluating if point belongs to the manifold.
        """

    @abc.abstractmethod
    def is_tangent(self, vector, base_point, atol=gs.atol):
        """Check whether the vector is tangent at base_point.

        Parameters
        ----------
        vector : array-like, shape=[..., *point_shape]
            Vector.
        base_point : array-like, shape=[..., *point_shape]
            Point on the manifold.
        atol : float
            Absolute tolerance.
            Optional, default: backend atol.

        Returns
        -------
        is_tangent : bool
            Boolean denoting if vector is a tangent vector at the base point.
        """

    @abc.abstractmethod
    def to_tangent(self, vector, base_point):
        """Project a vector to a tangent space of the manifold.

        Parameters
        ----------
        vector : array-like, shape=[..., *point_shape]
            Vector.
        base_point : array-like, shape=[..., *point_shape]
            Point on the manifold.

        Returns
        -------
        tangent_vec : array-like, shape=[..., *point_shape]
            Tangent vector at base point.
        """

    @abc.abstractmethod
    def random_point(self, n_samples=1, bound=1.0):
        """Sample random points on the manifold according to some distribution.

        If the manifold is compact, preferably a uniform distribution will be used.

        Parameters
        ----------
        n_samples : int
            Number of samples.
            Optional, default: 1.
        bound : float
            Bound of the interval in which to sample for non compact manifolds.
            Optional, default: 1.

        Returns
        -------
        samples : array-like, shape=[..., *point_shape]
            Points sampled on the manifold.
        """

    def regularize(self, point):
        """Regularize a point to the canonical representation for the manifold.

        Parameters
        ----------
        point : array-like, shape=[..., dim]
            Point.

        Returns
        -------
        regularized_point : array-like, shape=[..., *point_shape]
            Regularized point.
        """
        return gs.copy(point)

    def random_tangent_vec(self, base_point, n_samples=1):
        """Generate random tangent vec.

        Parameters
        ----------
        n_samples : int
            Number of samples.
            Optional, default: 1.
        base_point :  array-like, shape=[..., *point_shape]
            Point.

        Returns
        -------
        tangent_vec : array-like, shape=[..., *point_shape]
            Tangent vec at base point.
        """
        if (
            n_samples > 1
            and base_point.ndim > len(self.shape)
            and n_samples != len(base_point)
        ):
            raise ValueError(
                "The number of base points must be the same as the "
                "number of samples, when different from 1."
            )
        batch_size = () if n_samples == 1 else (n_samples,)
        vector = gs.random.normal(
            size=batch_size + self.shape, dtype=gs.get_default_cdtype()
        )
        return self.to_tangent(vector, base_point)

    def projection(self, point):
        """Project a point to the manifold.

        Parameters
        ----------
        point: array-like, shape[..., *point_shape]
            Point.

        Returns
        -------
        point: array-like, shape[..., *point_shape]
            Point.
        """
        if self.intrinsic:
            return gs.copy(point)

        raise NotImplementedError("`projection` is not implemented yet")



"""Base tools to handle datasets.

Note: deeply inspired by sklearn.
"""

import logging
import os
from collections import namedtuple
from urllib.request import urlretrieve

DEFAULT_DATA_DIR = os.path.expanduser(os.path.join("~", ".geomstats_data"))


RemoteFileMetadata = namedtuple("RemoteFileMetadata", ["filename", "url"])


def _get_data_home(data_home=None):
    if data_home is None:
        data_home = DEFAULT_DATA_DIR

    os.makedirs(data_home, exist_ok=True)

    return data_home


def _fetch_remote(url, filename, dirname=None):
    dirname = _get_data_home(data_home=dirname)

    file_path = os.path.join(dirname, filename)

    if os.path.exists(file_path):
        logging.info(
            f"Data has already been downloaded... using cached file ('{file_path}')."
        )
    else:
        logging.info(f"Downloading '{file_path}' from {url} to '{dirname}'.")
        urlretrieve(url, file_path)

    return file_path



import pytest

from geomstats.geometry.scalar_product_metric import (
    ScalarProductMetric,
    _ScaledMethodsRegistry,
    _wrap_attr,
    register_scaled_method,
)
from geomstats.test.test_case import TestCase


class WrapperTestCase(TestCase):
    def test_wrap_attr(self, func, scale):
        scaled_func = _wrap_attr(scale, func)

        res = func()
        scaled_res = scaled_func()

        self.assertAllClose(res, scaled_res / scale)

    def test_scaling_factor(self, func_name, scale, expected):
        scaling_factor = _ScaledMethodsRegistry._get_scaling_factor(func_name, scale)
        self.assertAllClose(scaling_factor, expected)

    def test_non_scaled(self, func_name, scale):
        scaling_factor = _ScaledMethodsRegistry._get_scaling_factor(func_name, scale)
        assert scaling_factor is None


class InstantiationTestCase(TestCase):
    @pytest.mark.random
    def test_scalar_metric_multiplication(self, scale):
        scaled_metric_1 = scale * self.space.metric
        scaled_metric_2 = self.space.metric * scale

        point_a, point_b = self.space.random_point(2)
        dist = self.space.metric.squared_dist(point_a, point_b)
        dist_1 = scaled_metric_1.squared_dist(point_a, point_b)
        dist_2 = scaled_metric_2.squared_dist(point_a, point_b)

        self.assertAllClose(scale * dist, dist_1)
        self.assertAllClose(scale * dist, dist_2)

        with pytest.raises(ValueError):
            scale * scaled_metric_1

    def test_scalar_metric_multiplication_error(self, scale):
        with pytest.raises(TypeError):
            scale * self.space.metric

        with pytest.raises(TypeError):
            self.space.metric * scale

        with pytest.raises(TypeError):
            self.space.metric * self.space.metric

    def test_scalar_metric_signature(self, scale):
        with pytest.raises(TypeError):
            ScalarProductMetric(self.space.metric, scale)

    @pytest.mark.random
    def test_scaling_scalar_metric(self, scale):
        point_a, point_b = self.space.random_point(2)

        dist = self.space.metric.squared_dist(point_a, point_b)

        self.space.equip_with_metric(ScalarProductMetric(self.space, scale))
        dist_1 = self.space.metric.squared_dist(point_a, point_b)

        self.space.equip_with_metric(ScalarProductMetric(self.space, scale))
        dist_2_a = self.space.metric.squared_dist(point_a, point_b)

        self.space.equip_with_metric().equip_with_metric(
            ScalarProductMetric(self.space, scale)
        ).equip_with_metric(scale * self.space.metric)
        dist_2_b = self.space.metric.squared_dist(point_a, point_b)

        self.space.equip_with_metric().equip_with_metric(
            ScalarProductMetric(self.space, scale)
        ).equip_with_metric(self.space.metric * scale)
        dist_2_c = self.space.metric.squared_dist(point_a, point_b)

        self.assertAllClose(scale * dist, dist_1)
        self.assertAllClose(scale**2 * dist, dist_2_a)
        self.assertAllClose(dist_2_b, dist_2_a)
        self.assertAllClose(dist_2_c, dist_2_a)


class CustomizationTestCase(TestCase):
    def test_register_scaled_method_invalid(self, func_name, scaling_type):
        with pytest.raises(ValueError):
            register_scaled_method(func_name, scaling_type)

    def test_register_scaled_method_valid(self, scale):
        def dummy_method(self):
            return 1

        self.Metric.dummy_method = dummy_method
        register_scaled_method("dummy_method", "linear")
        scaled_metric_1 = scale * self.space.metric
        self.assertAllClose(scale, scaled_metric_1.dummy_method())



"""The n-dimensional hyperbolic space.

The n-dimensional hyperbolic space embedded with
the hyperboloid representation (embedded in minkowsky space).

Lead author: Nina Miolane.
"""

import math

import geomstats.algebra_utils as utils
import geomstats.backend as gs
from geomstats.geometry._hyperbolic import _Hyperbolic
from geomstats.geometry.base import LevelSet
from geomstats.geometry.minkowski import Minkowski
from geomstats.geometry.riemannian_metric import RiemannianMetric
from geomstats.vectorization import repeat_out


class Hyperboloid(_Hyperbolic, LevelSet):
    """Class for the n-dimensional hyperboloid space.

    Class for the n-dimensional hyperboloid space as embedded in (n+1)-dimensional
    Minkowski space as the set of points with squared norm equal to -1, i.e.

    .. math::
        - x_0^2 + x_1^2 + ... + x_n^2 = - 1.

    For other
    representations of hyperbolic spaces see the `Hyperbolic` class.

    Parameters
    ----------
    dim : int
        Dimension of the hyperbolic space.
    """

    def __init__(self, dim, equip=True):
        self.coords_type = "extrinsic"
        self.dim = dim
        super().__init__(dim=dim, intrinsic=False, equip=equip)

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return HyperboloidMetric

    def _define_embedding_space(self):
        return Minkowski(self.dim + 1)

    def submersion(self, point):
        """Submersion that defines the manifold.

        Parameters
        ----------
        point : array-like, shape=[..., dim + 1]

        Returns
        -------
        submersed_point : array-like, shape=[...]
        """
        return self.embedding_space.metric.squared_norm(point) + 1.0

    def tangent_submersion(self, vector, point):
        """Tangent submersion.

        Parameters
        ----------
        vector : array-like, shape=[..., dim + 1]
        point : array-like, shape=[..., dim + 1]

        Returns
        -------
        submersed_vector : array-like, shape=[...]
        """
        return self.embedding_space.metric.inner_product(vector, point)

    def projection(self, point):
        """Project a point in space on the hyperboloid.

        Parameters
        ----------
        point : array-like, shape=[..., dim + 1]
            Point in embedding Euclidean space.

        Returns
        -------
        projected_point : array-like, shape=[..., dim + 1]
            Point projected on the hyperboloid.
        """
        belongs = self.belongs(point)

        # avoid dividing by 0
        factor = gs.where(point[..., 0] == 0.0, 1.0, point[..., 0] + gs.atol)

        first_coord = gs.where(belongs, 1.0, 1.0 / factor)
        intrinsic = gs.einsum("...,...i->...i", first_coord, point)[..., 1:]
        return self.intrinsic_to_extrinsic_coords(intrinsic)

    def regularize(self, point):
        """Regularize a point to the canonical representation.

        Regularize a point to the canonical representation chosen
        for the hyperbolic space, to avoid numerical issues.

        Parameters
        ----------
        point : array-like, shape=[..., dim + 1]
            Point.

        Returns
        -------
        projected_point : array-like, shape=[..., dim + 1]
            Point in hyperbolic space in canonical representation
            in extrinsic coordinates.
        """
        sq_norm = self.embedding_space.metric.squared_norm(point)
        if not gs.all(sq_norm):
            raise ValueError(
                "Cannot project a vector of norm 0. in the "
                "Minkowski space to the hyperboloid"
            )
        real_norm = gs.sqrt(gs.abs(sq_norm))
        return gs.einsum("...i,...->...i", point, 1.0 / real_norm)

    def to_tangent(self, vector, base_point):
        """Project a vector to a tangent space of the hyperbolic space.

        Project a vector in Minkowski space on the tangent space
        of the hyperbolic space at a base point.

        Parameters
        ----------
        vector : array-like, shape=[..., dim + 1]
            Vector in Minkowski space to be projected.
        base_point : array-like, shape=[..., dim + 1]
            Point in hyperbolic space.

        Returns
        -------
        tangent_vec : array-like, shape=[..., dim + 1]
            Tangent vector at the base point, equal to the projection of
            the vector in Minkowski space.
        """
        sq_norm = self.embedding_space.metric.squared_norm(base_point)
        inner_prod = self.embedding_space.metric.inner_product(base_point, vector)

        coef = inner_prod / sq_norm

        return vector - gs.einsum("...,...j->...j", coef, base_point)

    def intrinsic_to_extrinsic_coords(self, point_intrinsic):
        """Convert from intrinsic to extrinsic coordinates.

        Parameters
        ----------
        point_intrinsic : array-like, shape=[..., dim]
            Point in the embedded manifold in intrinsic coordinates.

        Returns
        -------
        point_extrinsic : array-like, shape=[..., dim + 1]
            Point in the embedded manifold in extrinsic coordinates.
        """
        return self.change_coordinates_system(point_intrinsic, "intrinsic", "extrinsic")

    def extrinsic_to_intrinsic_coords(self, point_extrinsic):
        """Convert from extrinsic to intrinsic coordinates.

        Parameters
        ----------
        point_extrinsic : array-like, shape=[..., dim + 1]
            Point in the embedded manifold in extrinsic coordinates,
            i. e. in the coordinates of the embedding manifold.

        Returns
        -------
        point_intrinsic : array-like, shape=[..., dim]
            Point in intrinsic coordinates.
        """
        return self.change_coordinates_system(point_extrinsic, "extrinsic", "intrinsic")

    def project_on_geodesic(self, point, base_point, tangent_vec):
        """Project on geodesic in extrinsic coordinates.

        Project point onto geodesic going through base point in direction
        of tangent vector. See reference below.

        Parameters
        ----------
        point: array-like, shape=[..., dim + 1]
            Point in hyperbolic space.
        base_point: array-like, shape=[..., dim + 1]
            Point through which the geodesic passes.
        tangent_vec : array-like, shape=[..., dim + 1]
            Tangent vector in Minkowski space, direction of the geodesic.

        Returns
        -------
        proj : array-like, shape=[..., dim + 1]
            Projected point on the geodesic.

        References
        ----------
        .. [CSV2016] R. Chakraborty, D. Seo, and B. C. Vemuri,
            "An efficient exact-pga algorithm for constant curvature manifolds."
            Proceedings of the IEEE conference on computer vision and pattern
            recognition. 2016.
        """
        inner_prod_1 = self.metric.inner_product(point, tangent_vec)
        inner_prod_2 = self.metric.inner_product(point, base_point)
        norm_v = self.metric.norm(tangent_vec, base_point)
        dist_to_proj = gs.arctanh(-inner_prod_1 / inner_prod_2 / norm_v)
        gs.einsum("...,...i->...i", gs.cosh(dist_to_proj), base_point)
        proj = gs.einsum(
            "...,...i->...i", gs.cosh(dist_to_proj), base_point
        ) + gs.einsum(
            "...,...i->...i",
            gs.sinh(dist_to_proj),
            gs.einsum("..., ...i ->...i", 1 / norm_v, tangent_vec),
        )
        return proj


class HyperboloidMetric(RiemannianMetric):
    """Class that defines operations using a hyperbolic metric."""

    def metric_matrix(self, base_point=None):
        """Compute the inner product matrix.

        Parameters
        ----------
        base_point: array-like, shape=[..., dim + 1]
            Base point.
            Optional, default: None.

        Returns
        -------
        inner_prod_mat: array-like, shape=[..., dim+1, dim + 1]
            Inner-product matrix.
        """
        return self._space.embedding_space.metric.metric_matrix(base_point)

    def inner_product(self, tangent_vec_a, tangent_vec_b, base_point=None):
        """Compute the inner-product of two tangent vectors at a base point.

        Parameters
        ----------
        tangent_vec_a : array-like, shape=[..., dim + 1]
            First tangent vector at base point.
        tangent_vec_b : array-like, shape=[..., dim + 1]
            Second tangent vector at base point.
        base_point : array-like, shape=[..., dim + 1], optional
            Point in hyperbolic space.

        Returns
        -------
        inner_prod : array-like, shape=[...,]
            Inner-product of the two tangent vectors.
        """
        return self._space.embedding_space.metric.inner_product(
            tangent_vec_a, tangent_vec_b, base_point
        )

    def squared_norm(self, vector, base_point=None):
        """Compute the squared norm of a vector.

        Squared norm of a vector associated with the inner-product
        at the tangent space at a base point.

        Parameters
        ----------
        vector : array-like, shape=[..., dim + 1]
            Vector on the tangent space of the hyperbolic space at base point.
        base_point : array-like, shape=[..., dim + 1], optional
            Point in hyperbolic space in extrinsic coordinates.

        Returns
        -------
        sq_norm : array-like, shape=[...,]
            Squared norm of the vector.
        """
        return self._space.embedding_space.metric.squared_norm(vector)

    def exp(self, tangent_vec, base_point):
        """Compute the Riemannian exponential of a tangent vector.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., dim + 1]
            Tangent vector at a base point.
        base_point : array-like, shape=[..., dim + 1]
            Point in hyperbolic space.

        Returns
        -------
        exp : array-like, shape=[..., dim + 1]
            Point in hyperbolic space equal to the Riemannian exponential
            of tangent_vec at the base point.
        """
        sq_norm_tangent_vec = self._space.embedding_space.metric.squared_norm(
            tangent_vec
        )
        sq_norm_tangent_vec = gs.clip(sq_norm_tangent_vec, 0, math.inf)

        coef_1 = utils.taylor_exp_even_func(
            sq_norm_tangent_vec, utils.cosh_close_0, order=5
        )
        coef_2 = utils.taylor_exp_even_func(
            sq_norm_tangent_vec, utils.sinch_close_0, order=5
        )

        exp = gs.einsum("...,...j->...j", coef_1, base_point) + gs.einsum(
            "...,...j->...j", coef_2, tangent_vec
        )

        return self._space.regularize(exp)

    def log(self, point, base_point):
        """Compute Riemannian logarithm of a point wrt a base point.

        If `coords_type` is `poincare` then base_point belongs
        to the Poincare ball and point is a vector in the Euclidean
        space of the same dimension as the ball.

        Parameters
        ----------
        point : array-like, shape=[..., dim + 1]
            Point in hyperbolic space.
        base_point : array-like, shape=[..., dim + 1]
            Point in hyperbolic space.

        Returns
        -------
        log : array-like, shape=[..., dim + 1]
            Tangent vector at the base point equal to the Riemannian logarithm
            of point at the base point.
        """
        angle = self.dist(base_point, point)

        coef_1_ = utils.taylor_exp_even_func(angle**2, utils.inv_sinch_close_0, order=4)
        coef_2_ = utils.taylor_exp_even_func(angle**2, utils.inv_tanh_close_0, order=4)

        log_term_1 = gs.einsum("...,...j->...j", coef_1_, point)
        log_term_2 = -gs.einsum("...,...j->...j", coef_2_, base_point)
        return log_term_1 + log_term_2

    def squared_dist(self, point_a, point_b):
        """Squared geodesic distance between two points.

        Parameters
        ----------
        point_a : array-like, shape=[..., dim]
            Point.
        point_b : array-like, shape=[..., dim]
            Point.

        Returns
        -------
        sq_dist : array-like, shape=[...,]
            Squared distance.
        """
        return self.dist(point_a, point_b) ** 2

    def dist(self, point_a, point_b):
        """Compute the geodesic distance between two points.

        Parameters
        ----------
        point_a : array-like, shape=[..., dim + 1]
            First point in hyperbolic space.
        point_b : array-like, shape=[..., dim + 1]
            Second point in hyperbolic space.

        Returns
        -------
        dist : array-like, shape=[...,]
            Geodesic distance between the two points.
        """
        embedding_metric = self._space.embedding_space.metric
        sq_norm_a = embedding_metric.squared_norm(point_a)
        sq_norm_b = embedding_metric.squared_norm(point_b)
        inner_prod = embedding_metric.inner_product(point_a, point_b)

        cosh_angle = -inner_prod / gs.sqrt(sq_norm_a * sq_norm_b)
        cosh_angle = gs.clip(cosh_angle, 1.0, 1e24)

        return gs.arccosh(cosh_angle)

    def parallel_transport(
        self, tangent_vec, base_point, direction=None, end_point=None
    ):
        r"""Compute the parallel transport of a tangent vector.

        Closed-form solution for the parallel transport of a tangent vector
        along the geodesic between two points `base_point` and `end_point`
        or alternatively defined by :math:`t \mapsto exp_{(base\_point)}(
        t*direction)`.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., dim + 1]
            Tangent vector at base point to be transported.
        base_point : array-like, shape=[..., dim + 1]
            Point on the hyperboloid.
        direction : array-like, shape=[..., dim + 1]
            Tangent vector at base point, along which the parallel transport
            is computed.
            Optional, default : None.
        end_point : array-like, shape=[..., dim + 1]
            Point on the hyperboloid. Point to transport to. Unused if `tangent_vec_b`
            is given.
            Optional, default : None.

        Returns
        -------
        transported_tangent_vec: array-like, shape=[..., dim + 1]
            Transported tangent vector at `exp_(base_point)(tangent_vec_b)`.
        """
        if direction is None:
            if end_point is not None:
                direction = self.log(end_point, base_point)
            else:
                raise ValueError(
                    "Either an end_point or a tangent_vec_b must be given to define the"
                    " geodesic along which to transport."
                )
        theta = self._space.embedding_space.metric.norm(direction)
        eps = gs.where(theta == 0.0, 1.0, theta)
        normalized_b = gs.einsum("...,...i->...i", 1 / eps, direction)
        pb = self._space.embedding_space.metric.inner_product(tangent_vec, normalized_b)
        p_orth = tangent_vec - gs.einsum("...,...i->...i", pb, normalized_b)
        transported = (
            gs.einsum("...,...i->...i", gs.sinh(theta) * pb, base_point)
            + gs.einsum("...,...i->...i", gs.cosh(theta) * pb, normalized_b)
            + p_orth
        )
        return transported

    def injectivity_radius(self, base_point=None):
        """Compute the radius of the injectivity domain.

        This is is the supremum of radii r for which the exponential map is a
        diffeomorphism from the open ball of radius r centered at the base
        point onto its image.
        In the case of the hyperbolic space, it does not depend on the base
        point and is infinite everywhere, because of the negative curvature.

        Parameters
        ----------
        base_point : array-like, shape=[..., dim+1]
            Point on the manifold.

        Returns
        -------
        radius : array-like, shape=[...,]
            Injectivity radius.
        """
        radius = gs.array(math.inf)
        return repeat_out(self._space.point_ndim, radius, base_point)



"""Base class for the n-dimensional hyperbolic space.

The n-dimensional hyperbolic regardless its different representations.
"""

import geomstats.backend as gs
from geomstats.geometry.diffeo import Diffeo


class HyperbolicDiffeo(Diffeo):
    """Diffeomorphism between hyperbolic coordinates.

    Parameters
    ----------
    from_coordinates_system : str, {'extrinsic', 'ball', 'half-space'}
        Coordinates type.
    to_coordinates_system : str, {'extrinsic', 'ball', 'half-space'}
        Coordinates type.
    """

    def __init__(self, from_coordinates_system, to_coordinates_system):
        self.from_coordinates_system = from_coordinates_system
        self.to_coordinates_system = to_coordinates_system

    def __call__(self, base_point):
        """Diffeomorphism at base point.

        Parameters
        ----------
        base_point : array-like, shape=[..., *space_shape]
            Base point.

        Returns
        -------
        image_point : array-like, shape=[..., *image_shape]
            Image point.
        """
        return _Hyperbolic.change_coordinates_system(
            base_point, self.from_coordinates_system, self.to_coordinates_system
        )

    def inverse(self, image_point):
        r"""Inverse diffeomorphism at image point.

        :math:`f^{-1}: N \rightarrow M`

        Parameters
        ----------
        image_point : array-like, shape=[..., *image_shape]
            Image point.

        Returns
        -------
        base_point : array-like, shape=[..., *space_shape]
            Base point.
        """
        return _Hyperbolic.change_coordinates_system(
            image_point, self.to_coordinates_system, self.from_coordinates_system
        )

    def tangent(self, tangent_vec, base_point=None, image_point=None):
        r"""Tangent diffeomorphism at base point.

        df_p is a linear map from T_pM to T_f(p)N.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., *space_shape]
            Tangent vector at base point.
        base_point : array-like, shape=[..., *space_shape]
            Base point.
        image_point : array-like, shape=[..., *image_shape]
            Image point.

        Returns
        -------
        image_tangent_vec : array-like, shape=[..., *image_shape]
            Image tangent vector at image of the base point.
        """
        if base_point is None:
            base_point = self.inverse(image_point)
        return _Hyperbolic.change_tangent_coordinates_system(
            tangent_vec,
            base_point,
            self.from_coordinates_system,
            self.to_coordinates_system,
        )

    def inverse_tangent(self, image_tangent_vec, image_point=None, base_point=None):
        r"""Inverse tangent diffeomorphism at image point.

        df^-1_p is a linear map from T_f(p)N to T_pM

        Parameters
        ----------
        image_tangent_vec : array-like, shape=[..., *image_shape]
            Image tangent vector at image point.
        image_point : array-like, shape=[..., *image_shape]
            Image point.
        base_point : array-like, shape=[..., *space_shape]
            Base point.

        Returns
        -------
        tangent_vec : array-like, shape=[..., *space_shape]
            Tangent vector at base point.
        """
        if image_point is None:
            image_point = self(base_point)

        return _Hyperbolic.change_tangent_coordinates_system(
            image_tangent_vec,
            image_point,
            self.to_coordinates_system,
            self.from_coordinates_system,
        )


def _scalarvecmul(scalar, vec):
    """Vectorized scalar vector multiplication.

    Parameters
    ----------
    scalar : array-like, shape=[...]
    vec : array-like, shape=[..., dim]
    """
    return gs.einsum("...,...i->...i", scalar, vec)


class _Hyperbolic:
    """Class for the n-dimensional hyperbolic space.

    Common class for the n-dimensional hyperbolic space regardless the
    representation being used. This class gathers the change of coordinate
    methods that are common to all representations.

    This class cannot be instantiated in itself, but a public class
    `Hyperbolic` is available, that returns the class that corresponds to the
    chosen representation: Hyperboloid, Poincare Ball or Poincare Half-space.
    """

    @staticmethod
    def _intrinsic_to_extrinsic_coordinates(point):
        """Convert intrinsic to extrinsic coordinates.

        Convert the parameterization of a point in hyperbolic space
        from its intrinsic coordinates, to its extrinsic coordinates
        in Minkowski space.

        Parameters
        ----------
        point : array-like, shape=[..., dim]
            Point in hyperbolic space in intrinsic coordinates.

        Returns
        -------
        point_extrinsic : array-like, shape=[..., dim + 1]
            Point in hyperbolic space in extrinsic coordinates.
        """
        coord_0 = gs.sqrt(1.0 + gs.sum(point**2, axis=-1))
        return gs.concatenate([coord_0[..., None], point], axis=-1)

    @staticmethod
    def _extrinsic_to_intrinsic_coordinates(point):
        """Convert extrinsic to intrinsic coordinates.

        Convert the parameterization of a point in hyperbolic space
        from its extrinsic coordinates in Minkowski space, to its
        intrinsic coordinates.

        Parameters
        ----------
        point : array-like, shape=[..., dim + 1]
            Point in hyperbolic space in extrinsic coordinates.

        Returns
        -------
        point_intrinsic : array-like, shape=[..., dim]
        """
        return point[..., 1:]

    @staticmethod
    def _extrinsic_to_ball_coordinates(point):
        """Convert extrinsic to ball coordinates.

        Convert the parameterization of a point in hyperbolic space
        from its intrinsic coordinates, to the Poincare ball model
        coordinates.

        Parameters
        ----------
        point : array-like, shape=[..., dim + 1]
            Point in hyperbolic space in extrinsic coordinates.

        Returns
        -------
        point_ball : array-like, shape=[..., dim]
            Point in hyperbolic space in Poincare ball coordinates.
        """
        return point[..., 1:] / (1 + point[..., :1])

    @staticmethod
    def _ball_to_extrinsic_coordinates(point):
        """Convert ball to extrinsic coordinates.

        Convert the parameterization of a point in hyperbolic space
        from its Poincare ball model coordinates, to the extrinsic
        coordinates.

        Parameters
        ----------
        point : array-like, shape=[..., dim]
            Point in hyperbolic space in Poincare ball coordinates.

        Returns
        -------
        point_extrinsic : array-like, shape=[..., dim + 1]
            Point in hyperbolic space in extrinsic coordinates.
        """
        squared_norm = gs.sum(point**2, -1)
        denominator = 1 - squared_norm
        t = (1 + squared_norm) / denominator
        intrinsic = _scalarvecmul(1.0 / denominator, 2 * point)
        return gs.concatenate([t[..., None], intrinsic], -1)

    @classmethod
    def _half_space_to_extrinsic_coordinates(cls, point):
        """Convert half-space to extrinsic coordinates.

        Convert the parameterization of a point in the hyperbolic space
        from its Poincare half-space coordinates, to the extrinsic
        coordinates.

        Parameters
        ----------
        point : array-like, shape=[..., dim]
            Point in hyperbolic space in half-space coordinates.

        Returns
        -------
        point_extrinsic : array-like, shape=[..., dim + 1]
            Point in hyperbolic space in extrinsic coordinates.
        """
        point_ball = cls.half_space_to_ball_coordinates(point)
        return cls._ball_to_extrinsic_coordinates(point_ball)

    @classmethod
    def _extrinsic_to_half_space_coordinates(cls, point):
        """Convert extrinsic to half-space coordinates.

        Convert the parameterization of a point in the hyperbolic space
        from its extrinsic coordinates, to the Poincare half-space
        coordinates.

        Parameters
        ----------
        point : array-like, shape=[..., dim + 1]
            Point in the hyperbolic space in extrinsic coordinates.

        Returns
        -------
        point_half_space : array-like, shape=[..., dim]
            Point in the hyperbolic space in half-space coordinates.
        """
        point_ball = cls._extrinsic_to_ball_coordinates(point)
        return cls.ball_to_half_space_coordinates(point_ball)

    @staticmethod
    def half_space_to_ball_coordinates(point):
        """Convert half-space to ball coordinates.

        Convert the parameterization of a point in the hyperbolic space
        from its Poincare half-space coordinates, to the Poincare ball
        coordinates.

        Parameters
        ----------
        point : array-like, shape=[..., dim]
            Point in the hyperbolic space in half-space coordinates.

        Returns
        -------
        point_ball : array-like, shape=[..., dim]
            Point in the hyperbolic space in Poincare ball coordinates.
        """
        sq_norm = gs.sum(point**2, axis=-1)
        den = 1 + sq_norm + 2 * point[..., -1]
        component_1 = _scalarvecmul(2.0 / den, point[..., :-1])
        component_2 = (sq_norm - 1) / den
        return gs.concatenate([component_1, component_2[..., None]], axis=-1)

    @staticmethod
    def ball_to_half_space_coordinates(point):
        """Convert ball to half space coordinates.

        Convert the parameterization of a point in the hyperbolic space
        from its Poincare ball coordinates, to the Poincare half-space
        coordinates.

        Parameters
        ----------
        point : array-like, shape=[..., dim]
            Point in the hyperbolic space in Poincare ball coordinates.

        Returns
        -------
        point_ball : array-like, shape=[..., dim]
            Point in the hyperbolic space in half-space coordinates.
        """
        sq_norm = gs.sum(point**2, axis=-1)
        den = 1 + sq_norm - 2 * point[..., -1]
        component_1 = _scalarvecmul(2.0 / den, point[..., :-1])
        component_2 = (1 - sq_norm) / den
        return gs.concatenate([component_1, component_2[..., None]], axis=-1)

    @staticmethod
    def half_space_to_ball_tangent(tangent_vec, base_point):
        """Convert half-space to ball tangent coordinates.

        Convert the parameterization of a tangent vector to the
        hyperbolic space from its Poincare half-space coordinates, to
        the Poinare ball coordinates.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., dim]
            Tangent vector at the base point in the Poincare half-space.
        base_point : array-like, shape=[..., dim]
            Point in the Poincare half-space.

        Returns
        -------
        tangent_vec_ball : array-like, shape=[..., dim]
           Tangent vector in the Poincare ball.
        """
        sq_norm = gs.sum(base_point**2, axis=-1)
        den = 1.0 + sq_norm + 2.0 * base_point[..., -1]
        scalar_prod = gs.sum(base_point * tangent_vec, axis=-1)
        component_1 = _scalarvecmul(
            2.0 / den, tangent_vec[..., :-1]
        ) - 4.0 * _scalarvecmul(
            (scalar_prod + tangent_vec[..., -1]) / den**2,
            base_point[..., :-1],
        )
        component_2 = (
            2 * scalar_prod / den
            - 2 * (sq_norm - 1) * (scalar_prod + tangent_vec[..., -1]) / den**2
        )
        tangent_vec_ball = gs.concatenate(
            [component_1, component_2[..., None]], axis=-1
        )

        return tangent_vec_ball

    @staticmethod
    def ball_to_half_space_tangent(tangent_vec, base_point):
        """Convert ball to half-space tangent coordinates.

        Convert the parameterization of a tangent vector to the
        hyperbolic space from its Poincare ball coordinates, to
        the Poinare half-space coordinates.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., dim]
            Tangent vector at the base point in the Poincare ball.
        base_point : array-like, shape=[..., dim]
            Point in the Poincare ball.

        Returns
        -------
        tangent_vec_half_space : array-like, shape=[..., dim]
            Tangent vector in the Poincare half-space.
        """
        sq_norm = gs.sum(base_point**2, axis=-1)
        den = 1 + sq_norm - 2 * base_point[..., -1]
        scalar_prod = gs.sum(base_point * tangent_vec, -1)
        component_1 = _scalarvecmul(
            2.0 / den, tangent_vec[..., :-1]
        ) - 4.0 * _scalarvecmul(
            (scalar_prod - tangent_vec[..., -1]) / den**2,
            base_point[..., :-1],
        )
        component_2 = (
            -2.0 * scalar_prod / den
            - 2 * (1.0 - sq_norm) * (scalar_prod - tangent_vec[..., -1]) / den**2
        )
        tangent_vec_half_space = gs.concatenate(
            [component_1, component_2[..., None]], axis=-1
        )

        return tangent_vec_half_space

    @staticmethod
    def extrinsic_to_ball_tangent(tangent_vec, base_point):
        """Convert extrinsic to ball tangent coordinates.

        Convert the parameterization of a tangent vector to the
        hyperbolic space from its extrinsic coordinates, to the
        Poincare ball coordinates.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., dim + 1]
            Tangent vector at the base point in extrinsic coordinates.
        base_point : array-like, shape=[..., dim + 1]
            Point in extrinsic coordinates.

        Returns
        -------
        tangent_vec_ball : array-like, shape=[..., dim]
            Tangent vector in the Poincare ball.
        """
        den = 1 + base_point[..., 0]
        return -_scalarvecmul(
            tangent_vec[..., 0] / den**2,
            base_point[..., 1:],
        ) + _scalarvecmul(1 / den, tangent_vec[..., 1:])

    @staticmethod
    def ball_to_extrinsic_tangent(tangent_vec, base_point):
        """Convert ball to extrinsic tangent coordinates.

        Convert the parameterization of a tangent vector to the
        hyperbolic space from its Poincare ball coordinates, to the
        extrinsic coordinates.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., dim]
            Tangent vector at the base point in the Poincare ball.
        base_point : array-like, shape=[..., dim]
            Point in the Poincare ball.

        Returns
        -------
        tangent_vec_extrinsic : array-like, shape=[..., dim + 1]
            Tangent vector in extrinsic coordinates.
        """
        sq_norm = gs.sum(base_point**2, axis=-1)
        scalar_prod = gs.sum(tangent_vec * base_point, axis=-1)
        den = (1 - sq_norm) ** 2

        term_1 = 2 * _scalarvecmul(1 - sq_norm, tangent_vec)
        term_2 = 4 * _scalarvecmul(scalar_prod, base_point)
        return _scalarvecmul(
            1 / den,
            gs.concatenate([4 * scalar_prod[..., None], term_1 + term_2], axis=-1),
        )

    @classmethod
    def half_space_to_extrinsic_tangent(cls, tangent_vec, base_point):
        """Convert half-space to extrinsinc tangent coordinates.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., dim]
            Tangent vector at the base point in the Poincare half-space.
        base_point : array-like, shape=[..., dim]
            Point in the Poincare half-space.

        Returns
        -------
        tangent_vec_extrinsic : array-like, shape=[..., dim + 1]
            Tangent vector in extrinsic coordinates.
        """
        tangent_vec_ball = cls.half_space_to_ball_tangent(tangent_vec, base_point)
        base_point_ball = cls.half_space_to_ball_coordinates(base_point)
        return cls.ball_to_extrinsic_tangent(tangent_vec_ball, base_point_ball)

    @classmethod
    def extrinsic_to_half_space_tangent(cls, tangent_vec, base_point):
        """Convert half-space to extrinsinc tangent coordinates.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., dim + 1]
            Tangent vector in extrinsic coordinates.
        base_point : array-like, shape=[..., dim + 1]
            Point in extrinsic coordinates.

        Returns
        -------
        tangent_vec_half_space : array-like, shape=[..., dim ]
            Tangent vector at the base point in the Poincare half-space.
        """
        tangent_vec_ball = cls.extrinsic_to_ball_tangent(tangent_vec, base_point)
        base_point_ball = cls._extrinsic_to_ball_coordinates(base_point)
        return cls.ball_to_half_space_tangent(tangent_vec_ball, base_point_ball)

    @staticmethod
    def change_coordinates_system(
        point, from_coordinates_system, to_coordinates_system
    ):
        """Convert coordinates of a point.

        Convert the parameterization of a point in the hyperbolic space
        from current given coordinate system to an other also given in
        parameters. The possible coordinates system are 'extrinsic',
        'intrinsic', 'ball' and 'half-plane' that correspond respectivelly
        to extrinsic coordinates in the hyperboloid , intrinsic
        coordinates in the hyperboloid, ball coordinates in the Poincare
        ball model and coordinates in the Poincare upper half-plane model.

        Parameters
        ----------
        point : array-like, shape=[..., {dim, dim + 1}]
            Point in hyperbolic space.
        from_coordinates_system : str, {'extrinsic', 'intrinsic', etc}
            Coordinates type.
        to_coordinates_system : str, {'extrinsic', 'intrinsic', etc}
            Coordinates type.

        Returns
        -------
        point_to : array-like, shape=[..., dim]
                               or shape=[n_sample, dim + 1]
            Point in hyperbolic space in coordinates given by
            to_coordinates_system.
        """
        coords_transform = {
            "ball-extrinsic": _Hyperbolic._ball_to_extrinsic_coordinates,
            "extrinsic-ball": _Hyperbolic._extrinsic_to_ball_coordinates,
            "ball-half-space": _Hyperbolic.ball_to_half_space_coordinates,
            "half-space-ball": _Hyperbolic.half_space_to_ball_coordinates,
            "intrinsic-extrinsic": _Hyperbolic._intrinsic_to_extrinsic_coordinates,
            "extrinsic-intrinsic": _Hyperbolic._extrinsic_to_intrinsic_coordinates,
            "extrinsic-half-space": _Hyperbolic._extrinsic_to_half_space_coordinates,
            "half-space-extrinsic": _Hyperbolic._half_space_to_extrinsic_coordinates,
        }

        if from_coordinates_system == to_coordinates_system:
            return gs.copy(point)

        func = coords_transform.get(
            f"{from_coordinates_system}-{to_coordinates_system}"
        )
        if func is not None:
            return func(point)

        extrinsic = coords_transform[f"{from_coordinates_system}-extrinsic"](point)
        return coords_transform[f"extrinsic-{to_coordinates_system}"](extrinsic)

    @staticmethod
    def change_tangent_coordinates_system(
        tangent_vec, base_point, from_coordinates_system, to_coordinates_system
    ):
        """Convert coordinates of a tangent vector.

        Convert the parameterization of a tangent vector in the hyperbolic space
        from current given coordinate system to an other also given in parameters.
         The possible coordinates system are 'extrinsic', 'ball' and
        'half-plane' that correspond respectively to extrinsic coordinates in the
        hyperboloid, coordinates in the Poincare ball model and coordinates in the
        Poincare upper half-space model.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., {dim, dim + 1}]
            Tangent vector in hyperbolic space.
        base_point : array-like, shape=[..., {dim, dim + 1}]
            Point in hyperbolic space.
        from_coordinates_system : str, {'extrinsic', 'intrinsic', etc}
            Coordinates type.
        to_coordinates_system : str, {'extrinsic', 'intrinsic', etc}
            Coordinates type.

        Returns
        -------
        tangent_vec_to : array-like, shape=[..., dim]
                               or shape=[n_sample, dim + 1]
            Tangent vector in hyperbolic space in coordinates given by
            to_coordinates_system.
        """
        coords_transform = {
            "ball-extrinsic": _Hyperbolic.ball_to_extrinsic_tangent,
            "extrinsic-ball": _Hyperbolic.extrinsic_to_ball_tangent,
            "ball-half-space": _Hyperbolic.ball_to_half_space_tangent,
            "half-space-ball": _Hyperbolic.half_space_to_ball_tangent,
            "extrinsic-half-space": _Hyperbolic.extrinsic_to_half_space_tangent,
            "half-space-extrinsic": _Hyperbolic.half_space_to_extrinsic_tangent,
        }

        if from_coordinates_system == to_coordinates_system:
            return gs.copy(tangent_vec)

        func = coords_transform.get(
            f"{from_coordinates_system}-{to_coordinates_system}"
        )
        return func(tangent_vec, base_point)

    def to_coordinates(self, point, to_coords_type="ball"):
        """Convert point to a target coordinate system.

        Convert the parameterization of a point in the hyperbolic space
        from current coordinate system to the coordinate system given.

        Parameters
        ----------
        point : array-like, shape=[..., {dim, dim + 1}]
            Point in hyperbolic space.
        to_coords_type : str, {'extrinsic', 'intrinsic', etc}
            Coordinates type.
            Optional, default: 'ball'.

        Returns
        -------
        point_to : array-like, shape=[..., {dim, dim + 1}]
            Point in hyperbolic space in coordinates given by to_coords_type.
        """
        return self.change_coordinates_system(point, self.coords_type, to_coords_type)

    def from_coordinates(self, point, from_coords_type):
        """Convert point to the current coordinate system.

        Convert the parameterization of a point in hyperbolic space
        from given coordinate system to the current coordinate system.

        Parameters
        ----------
        point : array-like, shape=[..., {dim, dim + 1}]
            Point in hyperbolic space in coordinates from_coords_type.
        from_coords_type : str, {'ball', 'extrinsic', 'intrinsic', ...}
            Coordinates type.

        Returns
        -------
        point_current : array-like, shape=[..., {dim, dim + 1}]
            Point in hyperbolic space.
        """
        return self.change_coordinates_system(point, from_coords_type, self.coords_type)

    def to_tangent_coordinates(self, tangent_vec, base_point, to_coords_type):
        """Convert tangent vector to a target coordinate system.

        Convert the parameterization of a tangent vector in the hyperbolic space
        from current coordinate system to the coordinate system given.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., {dim, dim + 1}]
            Tangent vector to hyperbolic space.
        base_point : array-like, shape=[..., {dim, dim + 1}]
            Point in hyperbolic space.
        to_coords_type : str, {'extrinsic', 'half-space', 'ball'}
            Coordinates type.

        Returns
        -------
        tangent_vec_to : array-like, shape=[..., {dim, dim + 1}]
            Tangent vector in hyperbolic space in coordinates given by to_coords_type.
        """
        return self.change_tangent_coordinates_system(
            tangent_vec, base_point, self.coords_type, to_coords_type
        )

    def from_tangent_coordinates(self, tangent_vec, base_point, from_coords_type):
        """Convert tangent vector to the current coordinate system.

        Convert the parameterization of a tangent vector to hyperbolic space
        from given coordinate system to the current coordinate system.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., {dim, dim + 1}]
            Tangent vector to hyperbolic space.
        base_point : array-like, shape=[..., {dim, dim + 1}]
            Point in hyperbolic space.
        from_coords_type : str, {'extrinsic', 'half-space', 'ball'}
            Coordinates type.

        Returns
        -------
        point_current : array-like, shape=[..., {dim, dim + 1}]
            Tangent vector in hyperbolic space.
        """
        return self.change_tangent_coordinates_system(
            tangent_vec, base_point, from_coords_type, self.coords_type
        )

    def random_point(self, n_samples=1, bound=1.0):
        """Sample over the hyperbolic space.

        Sample over the hyperbolic space. The sampling is performed by sampling from
        the uniform distribution with respect to the intrinsic co-ordinates. This is
        not uniform with respect to the volume measure. The function then transforms
        intrinsic samples into the selected co-ordinate system.

        Parameters
        ----------
        n_samples : int
            Number of samples.
            Optional, default: 1.
        bound: float
            Bound defining the hypersquare in which to sample.
            Optional, default: 1.

        Returns
        -------
        samples : array-like, shape=[..., dim + 1]
            Samples in hyperbolic space.
        """
        size = (n_samples, self.dim)
        samples = bound * 2.0 * (gs.random.rand(*size) - 0.5)

        samples = self.change_coordinates_system(samples, "intrinsic", self.coords_type)

        if n_samples == 1:
            return samples[0]
        return samples



"""Placeholders with error messages.

NumPy backend does not offer automatic differentiation.
The following functions return error messages.
"""

from geomstats.exceptions import AutodiffNotImplementedError

_USE_OTHER_BACKEND_MSG = (
    "Automatic differentiation is not supported with numpy backend. "
    "Use autograd or pytorch backend instead.\n"
    "Change backend via the command "
    "export GEOMSTATS_BACKEND=autograd in a terminal."
)


def value_and_grad(*args, **kwargs):
    """Return an error when using automatic differentiation with numpy."""
    raise AutodiffNotImplementedError(_USE_OTHER_BACKEND_MSG)


def value_and_jacobian(*args, **kwargs):
    """Return an error when using automatic differentiation with numpy."""
    raise AutodiffNotImplementedError(_USE_OTHER_BACKEND_MSG)


def jacobian(*args, **kwargs):
    """Return an error when using automatic differentiation with numpy."""
    raise AutodiffNotImplementedError(_USE_OTHER_BACKEND_MSG)


def jacobian_vec(*args, **kwargs):
    """Return an error when using automatic differentiation with numpy."""
    raise AutodiffNotImplementedError(_USE_OTHER_BACKEND_MSG)


def hessian(*args, **kwargs):
    """Return an error when using automatic differentiation with numpy."""
    raise AutodiffNotImplementedError(_USE_OTHER_BACKEND_MSG)


def hessian_vec(*args, **kwargs):
    """Return an error when using automatic differentiation with numpy."""
    raise AutodiffNotImplementedError(_USE_OTHER_BACKEND_MSG)


def jacobian_and_hessian(*args, **kwargs):
    """Return an error when using automatic differentiation with numpy."""
    raise AutodiffNotImplementedError(_USE_OTHER_BACKEND_MSG)


def custom_gradient(*grad_funcs):
    """Decorate a function to define its custom gradient(s).

    This is a placeholder in order to have consistent backend APIs.
    """

    def decorator(func):
        return func

    return decorator


def value_jacobian_and_hessian(*args, **kwargs):
    raise AutodiffNotImplementedError(_USE_OTHER_BACKEND_MSG)



"""Statistical Manifold of Gamma distributions with the Fisher metric.

The natural coordinate system for a Gamma Distribution is:
point = [kappa, nu], where kappa is the shape parameter, and nu the rate, or 1/scale.

However, information geometry most often works with standard coordinates, given by:
point = [kappa, gamma] = [kappa, kappa/nu].

The standard coordinate system is the convention we use in this script.
All points and all vectors input are assumed to be given in the standard coordinate
system unless stated otherwise.

Some of the methods in GammaDistributions allow to easily make the associated
change of variable, either for a point or a vector.

Lead author: Jules Deschamps.
"""

from scipy.stats import gamma

import geomstats.backend as gs
from geomstats.algebra_utils import from_vector_to_diagonal_matrix
from geomstats.geometry.base import VectorSpaceOpenSet
from geomstats.geometry.diffeo import InvolutionDiffeomorphism
from geomstats.geometry.euclidean import Euclidean
from geomstats.geometry.riemannian_metric import RiemannianMetric
from geomstats.information_geometry.base import (
    InformationManifoldMixin,
    ScipyUnivariateRandomVariable,
)
from geomstats.numerics.bvp import ScipySolveBVP
from geomstats.numerics.geodesic import ExpODESolver, LogODESolver
from geomstats.numerics.ivp import ScipySolveIVP
from geomstats.vectorization import get_batch_shape


class NaturalToStandardDiffeo(InvolutionDiffeomorphism):
    """Diffeomorphism between natural and standard coordinates."""

    def __call__(self, point):
        """Convert point from natural coordinates to standard coordinates.

        The change of variable is symmetric.

        Parameters
        ----------
        point : array-like, shape=[..., 2]
            Point of the Gamma manifold, given in natural coordinates.

        Returns
        -------
        point : array-like, shape=[..., 2]
            Point of the Gamma manifold, given in standard coordinates.
        """
        return gs.stack([point[..., 0], point[..., 0] / point[..., 1]], axis=-1)

    def tangent(self, tangent_vec, base_point=None, image_point=None):
        """Convert tangent vector from natural coordinates to standard coordinates.

        The change of variable is symmetric.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., 2]
            Tangent vector at base_point, given in natural coordinates.
        base_point : array-like, shape=[..., 2]
            Point of the Gamma manifold, given in natural coordinates.
        image_point : array-like, shape=[..., 2]
            Point of the Gamma manifold, given in standard coordinates.

        Returns
        -------
        image_tangent_vec : array-like, shape=[..., 2]
            Tangent vector at base_point, given in standard coordinates.
        """
        if base_point is None:
            base_point = self(image_point)

        kappa, scale = base_point[..., 0], base_point[..., 1]

        jac_row_1 = gs.array([1.0, 0.0])
        jac_row_2 = gs.stack([1 / scale, -kappa / scale**2], axis=-1)

        point_batch_shape = get_batch_shape(1, base_point)
        if point_batch_shape:
            jac_row_1 = gs.broadcast_to(jac_row_1, point_batch_shape + (2,))

        jac = gs.stack([jac_row_1, jac_row_2], axis=-2)

        return gs.einsum("...jk,...k->...j", jac, tangent_vec)


class GammaDistributions(InformationManifoldMixin, VectorSpaceOpenSet):
    """Class for the manifold of Gamma distributions.

    This is :math:`Gamma = (R_+^*)^2`, the positive quadrant of the
    2-dimensional Euclidean space.
    """

    def __init__(self, equip=True):
        super().__init__(
            dim=2,
            embedding_space=Euclidean(2, equip=False),
            support_shape=(),
            equip=equip,
        )

        self._scp_rv = GammaDistributionsRandomVariable(self)

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return GammaMetric

    def belongs(self, point, atol=gs.atol):
        """Evaluate if a point belongs to the manifold of Gamma distributions.

        Check that point defines parameters for a Gamma distribution,
        i.e. belongs to the positive quadrant of the Euclidean space.

        Parameters
        ----------
        point : array-like, shape=[..., 2]
            Point to be checked.
        atol : float
            Tolerance to evaluate positivity.
            Optional, default: gs.atol

        Returns
        -------
        belongs : array-like, shape=[...,]
            Boolean indicating whether point represents a Gamma
            distribution.
        """
        point_dim = point.shape[-1]
        belongs = point_dim == 2
        return gs.logical_and(belongs, gs.all(point >= -atol, axis=-1))

    def random_point(self, n_samples=1, upper_bound=5.0, lower_bound=0.0):
        """Sample parameters of Gamma distributions.

        The uniform distribution on [0, bound]^2 is used.

        Parameters
        ----------
        n_samples : int
            Number of samples.
            Optional, default: 1.
        bound : float
            Side of the square where the Gamma parameters are sampled.
            Optional, default: 5.

        Returns
        -------
        samples : array-like, shape=[..., 2]
            Sample of points representing Gamma distributions.
        """
        upper_bound = upper_bound * gs.ones(2)
        lower_bound = lower_bound * gs.ones(2)

        if gs.any((upper_bound - lower_bound) < 0):
            raise ValueError("upper_bound cannot be greater than lower_bound.")

        size = (2,) if n_samples == 1 else (n_samples, 2)
        return lower_bound + (upper_bound - lower_bound) * gs.random.rand(*size)

    def projection(self, point, atol=gs.atol):
        """Project a point in ambient space to the open set.

        The last coordinate is floored to `gs.atol` if it is negative.

        Parameters
        ----------
        point : array-like, shape=[..., 2]
            Point in ambient space.
        atol : float
            Tolerance to evaluate positivity.

        Returns
        -------
        projected : array-like, shape=[..., 2]
            Projected point.
        """
        return gs.where(point < atol, atol, point)

    def sample(self, point, n_samples=1):
        """Sample from the Gamma distribution.

        Sample from the Gamma distribution with parameters provided
        by point. This gives n_samples points.

        Parameters
        ----------
        point : array-like, shape=[..., dim]
            Point representing a Gamma distribution.
        n_samples : int
            Number of points to sample for each set of parameters in point.
            Optional, default: 1.

        Returns
        -------
        samples : array-like, shape=[..., n_samples]
            Sample from the Gamma distributions.
        """
        return self._scp_rv.rvs(point, n_samples)

    def point_to_pdf(self, point):
        """Compute pdf associated to point.

        Compute the probability density function of the Gamma
        distribution with parameters provided by point.

        Parameters
        ----------
        point : array-like, shape=[..., dim]
            Point representing a Gamma distribution.

        Returns
        -------
        pdf : function
            Probability density function of the Gamma distribution with
            parameters provided by point.
        """
        kappa = gs.expand_dims(point[..., 0], axis=-1)
        gamma = gs.expand_dims(point[..., 1], axis=-1)

        def pdf(x):
            """Generate parameterized function for Gamma pdf.

            Parameters
            ----------
            x : array-like, shape=[n_samples,]
                Points at which to compute the probability
                density function.

            Returns
            -------
            pdf_at_x : array-like, shape=[..., n_samples]
                Values of pdf at x for each value of the parameters provided
                by point.
            """
            x = gs.reshape(gs.array(x), (-1,))
            return (
                kappa**kappa
                * x ** (kappa - 1)
                * gs.exp(-kappa * x / gamma)
                / (gamma**kappa * gs.gamma(kappa))
            )

        return pdf

    @staticmethod
    def maximum_likelihood_fit(data):
        """Estimate parameters from samples.

        This is a wrapper around scipy's maximum likelihood estimator to
        estimate the parameters of a gamma distribution from samples.

        Parameters
        ----------
        data : list or list of lists/arrays
            Data to estimate parameters from. Lists of
            different length may be passed.

        Returns
        -------
        parameter : array-like, shape=[..., 2]
            Estimate of parameter obtained by maximum likelihood.
        """

        def is_nested(sample):
            """Check if sample contains an iterable."""
            for el in sample:
                try:
                    return iter(el)
                except TypeError:
                    return False

        if not is_nested(data):
            data = [data]
        parameters = []
        for sample in data:
            sample = gs.array(sample)
            kappa, _, scale = gamma.fit(sample, floc=0)
            nu = 1 / scale
            parameters.append(gs.array([kappa, kappa / nu]))
        return parameters[0] if len(data) == 1 else gs.stack(parameters)


class GammaMetric(RiemannianMetric):
    """Class for the Fisher information metric on Gamma distributions.

    References
    ----------
    .. [AD2008] Arwini, K. A., & Dodson, C. T. (2008).
        Information geometry (pp. 31-54). Springer Berlin Heidelberg.
    """

    def __init__(self, space):
        super().__init__(space=space)

        self.log_solver = LogODESolver(
            space, n_nodes=1000, integrator=ScipySolveBVP(max_nodes=1000)
        )
        self.exp_solver = ExpODESolver(space, integrator=ScipySolveIVP(method="LSODA"))

    def metric_matrix(self, base_point):
        """Compute the inner-product matrix.

        Compute the inner-product matrix of the Fisher information metric
        at the tangent space at base point.

        Parameters
        ----------
        base_point : array-like, shape=[..., 2]
            Base point.

        Returns
        -------
        mat : array-like, shape=[..., 2, 2]
            Inner-product matrix.
        """
        kappa, gamma = base_point[..., 0], base_point[..., 1]
        mat_diag = gs.stack(
            [gs.polygamma(1, kappa) - 1 / kappa, kappa / gamma**2], axis=-1
        )
        return from_vector_to_diagonal_matrix(mat_diag)

    def christoffels(self, base_point):
        """Compute the Christoffel symbols.

        Compute the Christoffel symbols of the Fisher information metric.
        For computation purposes, we replace the value of
        (gs.polygamma(1, x) - 1/x) by an equivalent (close lower-bound) when it becomes
        too difficult to compute, as per in [GQ2015]_.

        Parameters
        ----------
        base_point : array-like, shape=[..., 2]
            Base point.

        Returns
        -------
        christoffels : array-like, shape=[..., 2, 2, 2]
            Christoffel symbols, with the contravariant index on
            the first dimension.
            :math:`christoffels[..., i, j, k] = Gamma^i_{jk}`

        References
        ----------
        .. [AD2008] Arwini, K. A., & Dodson, C. T. (2008).
            Information geometry (pp. 31-54). Springer Berlin Heidelberg.

        .. [GQ2015] Guo, B. N., Qi, F., Zhao, J. L., & Luo, Q. M. (2015).
            Sharp inequalities for polygamma functions.
            Mathematica Slovaca, 65(1), 103-120.
        """
        base_point = gs.to_ndarray(base_point, to_ndim=2)

        kappa, gamma = base_point[:, 0], base_point[:, 1]

        if gs.any(kappa > 4e15):
            raise ValueError(
                "Christoffels computation overflows with values of kappa. "
                "All values of kappa < 4e15 work."
            )

        shape = kappa.shape

        c111 = gs.where(
            gs.polygamma(1, kappa) - 1 / kappa > gs.atol,
            (gs.polygamma(2, kappa) + gs.array(kappa) ** -2)
            / (2 * (gs.polygamma(1, kappa) - 1 / kappa)),
            0.25 * (kappa**2 * gs.polygamma(2, kappa) + 1),
        )

        c122 = gs.where(
            gs.polygamma(1, kappa) - 1 / kappa > gs.atol,
            -1 / (2 * gamma**2 * (gs.polygamma(1, kappa) - 1 / kappa)),
            -(kappa**2) / (4 * gamma**2),
        )

        c1 = gs.squeeze(
            from_vector_to_diagonal_matrix(gs.transpose(gs.array([c111, c122])))
        )

        c2 = gs.squeeze(
            gs.transpose(
                gs.array(
                    [[gs.zeros(shape), 1 / (2 * kappa)], [1 / (2 * kappa), -1 / gamma]]
                )
            )
        )

        christoffels = gs.array([c1, c2])

        if len(christoffels.shape) == 4:
            christoffels = gs.transpose(christoffels, [1, 0, 2, 3])

        return gs.squeeze(christoffels)

    def jacobian_christoffels(self, base_point):
        """Compute the Jacobian of the Christoffel symbols.

        Compute the Jacobian of the Christoffel symbols of the
        Fisher information metric.

        For computation purposes, we replace the value of
        (gs.polygamma(1, x) - 1/x) and (gs.polygamma(2,x) + 1/x**2) by an equivalent
        (close bounds) when they become too difficult to compute.

        Parameters
        ----------
        base_point : array-like, shape=[..., 2]
            Base point.

        Returns
        -------
        jac : array-like, shape=[..., 2, 2, 2, 2]
            Jacobian of the Christoffel symbols.
            :math:`jac[..., i, j, k, l] = dGamma^i_{jk} / dx_l`

        References
        ----------
        .. [GQ2015] Guo, B. N., Qi, F., Zhao, J. L., & Luo, Q. M. (2015).
            Sharp inequalities for polygamma functions.
            Mathematica Slovaca, 65(1), 103-120.
        """
        base_point = gs.to_ndarray(base_point, 2)

        n_points = base_point.shape[0]

        kappa, gamma = base_point[:, 0], base_point[:, 1]

        term_0 = gs.zeros((n_points))
        term_1 = 1 / gamma**2
        term_2 = gs.where(
            gs.polygamma(1, kappa) - 1 / kappa > gs.atol,
            kappa / (gamma**3 * (kappa * gs.polygamma(1, kappa) - 1)),
            kappa**2 / gamma**3,
        )
        term_3 = -1 / (2 * kappa**2)
        term_4 = gs.where(
            gs.polygamma(1, kappa) - 1 / kappa > gs.atol,
            (kappa**2 * gs.polygamma(2, kappa) + 1)
            / (2 * gamma**2 * (kappa * gs.polygamma(1, kappa) - 1) ** 2),
            (kappa**4 * gs.polygamma(2, kappa) + kappa**2) / (2 * gamma**2),
        )
        term_5 = gs.where(
            gs.polygamma(1, kappa) - 1 / kappa > gs.atol,
            (
                kappa**4
                * (
                    gs.polygamma(1, kappa) * gs.polygamma(3, kappa)
                    - gs.polygamma(2, kappa) ** 2
                )
                - kappa**3 * gs.polygamma(3, kappa)
                - 2 * kappa**2 * gs.polygamma(2, kappa)
                - 2 * kappa * gs.polygamma(1, kappa)
                + 1
            )
            / (2 * (kappa**2 * gs.polygamma(1, kappa) - kappa) ** 2),
            0.5
            * (
                kappa**4
                * (
                    gs.polygamma(1, kappa) * gs.polygamma(3, kappa)
                    - gs.polygamma(2, kappa) ** 2
                )
                - kappa**3 * gs.polygamma(3, kappa)
                - 2 * kappa**2 * gs.polygamma(2, kappa)
                - 2 * kappa * gs.polygamma(1, kappa)
                + 1
            ),
        )

        jac = gs.array(
            [
                [
                    [[term_5, term_0], [term_0, term_0]],
                    [[term_0, term_0], [term_4, term_2]],
                ],
                [
                    [[term_0, term_0], [term_3, term_0]],
                    [[term_3, term_0], [term_0, term_1]],
                ],
            ]
        )

        if n_points > 1:
            jac = gs.transpose(jac, [4, 0, 1, 2, 3])

        return gs.squeeze(jac)


class GammaDistributionsRandomVariable(ScipyUnivariateRandomVariable):
    """A gamma random variable."""

    def __init__(self, space):
        super().__init__(space, gamma.rvs, gamma.pdf)

    @staticmethod
    def _flatten_params(point, pre_flat_shape):
        param_a = gs.expand_dims(point[..., 0], axis=-1)
        scale = gs.expand_dims(point[..., 1] / point[..., 0], axis=-1)

        flat_param_a = gs.reshape(gs.broadcast_to(param_a, pre_flat_shape), (-1,))
        flat_scale = gs.reshape(gs.broadcast_to(scale, pre_flat_shape), (-1,))
        return {"a": flat_param_a, "scale": flat_scale}



"""The manifold of Hermitian positive definite (HPD) matrices.

Lead author: Yann Cabanes.


References
----------
.. [Cabanes2022] Yann Cabanes. Multidimensional complex stationary
    centered Gaussian autoregressive time series machine learning
    in Poincaré and Siegel disks: application for audio and radar
    clutter classification, PhD thesis, 2022
.. [JV2016] B. Jeuris and R. Vandebril. The Kahler mean of Block-Toeplitz
    matrices with Toeplitz structured blocks, 2016.
    https://epubs.siam.org/doi/pdf/10.1137/15M102112X
"""

import math

import geomstats.backend as gs
from geomstats.geometry.base import ComplexVectorSpaceOpenSet
from geomstats.geometry.complex_matrices import ComplexMatrices, ComplexMatricesMetric
from geomstats.geometry.complex_riemannian_metric import ComplexRiemannianMetric
from geomstats.geometry.general_linear import GeneralLinear
from geomstats.geometry.hermitian_matrices import HermitianMatrices, expmh, powermh
from geomstats.geometry.matrices import Matrices
from geomstats.geometry.pullback_metric import PullbackDiffeoMetric
from geomstats.geometry.spd_matrices import SymMatrixLog, logmh
from geomstats.integrator import integrate
from geomstats.vectorization import repeat_out


class HPDMatrices(ComplexVectorSpaceOpenSet):
    """Class for the manifold of Hermitian positive definite (HPD) matrices.

    Parameters
    ----------
    n : int
        Integer representing the shape of the matrices: n x n.
    """

    def __init__(self, n, equip=True):
        super().__init__(
            dim=n**2, shape=(n, n), embedding_space=HermitianMatrices(n), equip=equip
        )
        self.n = n

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return HPDAffineMetric

    @staticmethod
    def belongs(point, atol=gs.atol):
        """Check if a matrix is Hermitian with positive eigenvalues.

        Parameters
        ----------
        point : array-like, shape=[..., n, n]
            Point to be checked.
        atol : float
            Tolerance.
            Optional, default: backend atol.

        Returns
        -------
        belongs : array-like, shape=[...,]
            Boolean denoting if mat is an HPD matrix.
        """
        return ComplexMatrices.is_hpd(point, atol)

    def projection(self, point, atol=gs.atol):
        """Project a matrix to the space of HPD matrices.

        First the Hermitian part of point is computed, then the eigenvalues
        are floored to gs.atol.

        Parameters
        ----------
        point : array-like, shape=[..., n, n]
            Matrix to project.
        atol : float
            Tolerance.
            Optional, default: backend atol.

        Returns
        -------
        projected: array-like, shape=[..., n, n]
            HPD matrix.
        """
        herm = ComplexMatrices.to_hermitian(point)
        eigvals, eigvecs = gs.linalg.eigh(herm)
        regularized = gs.where(eigvals < atol, atol, eigvals)
        reconstruction = gs.einsum("...ij,...j->...ij", eigvecs, regularized)
        return Matrices.mul(reconstruction, ComplexMatrices.transconjugate(eigvecs))

    def random_point(self, n_samples=1, bound=0.1):
        """Sample in HPD(n) from the log-uniform distribution.

        Parameters
        ----------
        n_samples : int
            Number of samples.
            Optional, default: 1.
        bound : float
            Bound of the interval in which to sample in the tangent space.
            Optional, default: 1.

        Returns
        -------
        samples : array-like, shape=[..., n, n]
            Points sampled in HPD(n).
        """
        n = self.n
        size = (n_samples, n, n) if n_samples != 1 else (n, n)
        eye = gs.eye(n, dtype=gs.get_default_cdtype())
        samples = gs.stack([eye for i_sample in range(n_samples)], axis=0)
        samples = gs.reshape(samples, size)
        samples += bound * gs.random.rand(*size, dtype=gs.get_default_cdtype())
        samples = self.projection(samples)
        return samples

    def random_tangent_vec(self, base_point, n_samples=1):
        """Sample on the tangent space of HPD(n) from the uniform distribution.

        Parameters
        ----------
        n_samples : int
            Number of samples.
            Optional, default: 1.
        base_point : array-like, shape=[..., n, n]
            Base point of the tangent space.

        Returns
        -------
        samples : array-like, shape=[..., n, n]
            Points sampled in the tangent space at base_point.
        """
        n = self.n
        size = (n_samples, n, n) if n_samples != 1 else (n, n)

        sqrt_base_point = gs.cast(
            gs.linalg.sqrtm(base_point),
            base_point.dtype,
        )

        tangent_vec_at_id_aux = gs.random.rand(*size, dtype=gs.get_default_cdtype())
        tangent_vec_at_id_aux *= 2
        tangent_vec_at_id_aux -= 1 + 1j
        tangent_vec_at_id = tangent_vec_at_id_aux + ComplexMatrices.transconjugate(
            tangent_vec_at_id_aux
        )

        return Matrices.mul(sqrt_base_point, tangent_vec_at_id, sqrt_base_point)


class HPDAffineMetric(ComplexRiemannianMetric):
    """Class for the affine-invariant metric on the HPD manifold."""

    def inner_product(self, tangent_vec_a, tangent_vec_b, base_point):
        """Compute the affine-invariant inner-product.

        Compute the inner-product of tangent_vec_a and tangent_vec_b
        at point base_point using the affine invariant Riemannian metric.

        Parameters
        ----------
        tangent_vec_a : array-like, shape=[..., n, n]
            Tangent vector at base point.
        tangent_vec_b : array-like, shape=[..., n, n]
            Tangent vector at base point.
        base_point : array-like, shape=[..., n, n]
            Base point.

        Returns
        -------
        inner_product : array-like, shape=[..., n, n]
            Inner-product.
        """
        inv_base_point = GeneralLinear.inverse(base_point)
        aux_a = Matrices.mul(inv_base_point, tangent_vec_a)
        aux_b = Matrices.mul(inv_base_point, tangent_vec_b)

        return Matrices.trace_product(aux_a, aux_b)

    def exp(self, tangent_vec, base_point):
        """Compute the affine-invariant exponential map.

        Compute the Riemannian exponential at point base_point
        of tangent vector tangent_vec wrt the metric defined in inner_product.
        This gives a Hermitian positive definite matrix.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., n, n]
            Tangent vector at base point.
        base_point : array-like, shape=[..., n, n]
            Base point.

        Returns
        -------
        exp : array-like, shape=[..., n, n]
            Riemannian exponential.
        """
        sqrt_base_point, inv_sqrt_base_point = powermh(base_point, [1.0 / 2, -1.0 / 2])

        tangent_vec_at_id = Matrices.mul(
            inv_sqrt_base_point, tangent_vec, inv_sqrt_base_point
        )

        tangent_vec_at_id = ComplexMatrices.to_hermitian(tangent_vec_at_id)
        exp_from_id = expmh(tangent_vec_at_id)

        return Matrices.mul(sqrt_base_point, exp_from_id, sqrt_base_point)

    def log(self, point, base_point):
        """Compute the affine-invariant logarithm map.

        Compute the Riemannian logarithm at point base_point,
        of point wrt the metric defined in inner_product.
        This gives a tangent vector at point base_point.

        Parameters
        ----------
        point : array-like, shape=[..., n, n]
            Point.
        base_point : array-like, shape=[..., n, n]
            Base point.

        Returns
        -------
        log : array-like, shape=[..., n, n]
            Riemannian logarithm of point at base_point.
        """
        sqrt_base_point, inv_sqrt_base_point = powermh(base_point, [1.0 / 2, -1.0 / 2])
        point_near_id = Matrices.mul(inv_sqrt_base_point, point, inv_sqrt_base_point)

        # TODO: only this differs
        point_near_id = ComplexMatrices.to_hermitian(point_near_id)

        log_at_id = logmh(point_near_id)
        return Matrices.mul(sqrt_base_point, log_at_id, sqrt_base_point)

    def parallel_transport(
        self, tangent_vec, base_point, direction=None, end_point=None
    ):
        r"""Parallel transport of a tangent vector.

        Closed-form solution for the parallel transport of a tangent vector
        along the geodesic between two points `base_point` and `end_point`
        or alternatively defined by :math:`t \mapsto exp_{(base\_point)}(
        t*direction)`.
        Denoting `tangent_vec_a` by `S`, `base_point` by `A`, and `end_point`
        by `B` or `B = Exp_A(tangent_vec_b)` and :math:`E = (BA^{- 1})^{( 1
        / 2)}`. Then the parallel transport to `B` is:

        .. math::
            S' = ESE^T

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., n, n]
            Tangent vector at base point to be transported.
        base_point : array-like, shape=[..., n, n]
            Point on the manifold of HPD matrices. Point to transport from
        direction : array-like, shape=[..., n, n]
            Tangent vector at base point, initial speed of the geodesic along
            which the parallel transport is computed. Unused if `end_point` is given.
            Optional, default: None.
        end_point : array-like, shape=[..., n, n]
            Point on the manifold of HPD matrices. Point to transport to.
            Optional, default: None.

        Returns
        -------
        transported_tangent_vec: array-like, shape=[..., n, n]
            Transported tangent vector at exp_(base_point)(tangent_vec_b).
        """
        if end_point is None:
            end_point = self.exp(direction, base_point)
        sqrt_bp, inv_sqrt_bp = powermh(base_point, [1.0 / 2, -1.0 / 2])
        pdt = powermh(Matrices.mul(inv_sqrt_bp, end_point, inv_sqrt_bp), 1.0 / 2)
        congruence_mat = Matrices.mul(sqrt_bp, pdt, inv_sqrt_bp)
        return ComplexMatrices.congruent(tangent_vec, congruence_mat)

    def injectivity_radius(self, base_point=None):
        """Radius of the largest ball where the exponential is injective.

        Because of the negative curvature of this space, the injectivity radius
        is infinite everywhere.

        Parameters
        ----------
        base_point : array-like, shape=[..., n, n]
            Point on the manifold.

        Returns
        -------
        radius : array-like, shape=[...,]
            Injectivity radius.
        """
        radius = gs.array(math.inf)
        return repeat_out(self._space.point_ndim, radius, base_point)


class HPDBuresWassersteinMetric(ComplexRiemannianMetric):
    """Class for the Bures-Wasserstein metric on the HPD manifold."""

    def inner_product(self, tangent_vec_a, tangent_vec_b, base_point):
        r"""Compute the Bures-Wasserstein inner-product.

        Compute the inner-product of tangent_vec_a :math:`A` and tangent_vec_b
        :math:`B` at point base_point :math:`S=PDP^\top` using the
        Bures-Wasserstein Riemannian metric:

        .. math::
            \frac{1}{2}\sum_{i,j}\frac{[P^\top AP]_{ij}[P^\top BP]_{ij}}{d_i+d_j}

        Parameters
        ----------
        tangent_vec_a : array-like, shape=[..., n, n]
            Tangent vector at base point.
        tangent_vec_b : array-like, shape=[..., n, n]
            Tangent vector at base point.
        base_point : array-like, shape=[..., n, n]
            Base point.

        Returns
        -------
        inner_product : array-like, shape=[...,]
            Inner-product.
        """
        eigvals, eigvecs = gs.linalg.eigh(base_point)
        transconj_eigvecs = ComplexMatrices.transconjugate(eigvecs)
        rotated_tangent_vec_a = Matrices.mul(transconj_eigvecs, tangent_vec_a, eigvecs)
        rotated_tangent_vec_b = Matrices.mul(transconj_eigvecs, tangent_vec_b, eigvecs)

        coefficients = 1 / (eigvals[..., :, None] + eigvals[..., None, :])
        result = (
            ComplexMatrices.frobenius_product(
                gs.cast(coefficients, dtype=rotated_tangent_vec_a.dtype)
                * rotated_tangent_vec_a,
                rotated_tangent_vec_b,
            )
            / 2
        )

        return result

    def exp(self, tangent_vec, base_point):
        """Compute the Bures-Wasserstein exponential map.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., n, n]
            Tangent vector at base point.
        base_point : array-like, shape=[..., n, n]
            Base point.

        Returns
        -------
        exp : array-like, shape=[...,]
            Riemannian exponential.
        """
        eigvals, eigvecs = gs.linalg.eigh(base_point)
        transconj_eigvecs = ComplexMatrices.transconjugate(eigvecs)
        rotated_tangent_vec = Matrices.mul(transconj_eigvecs, tangent_vec, eigvecs)
        coefficients = 1 / (eigvals[..., :, None] + eigvals[..., None, :])
        rotated_sylvester = rotated_tangent_vec * gs.cast(
            coefficients, dtype=rotated_tangent_vec.dtype
        )
        rotated_hessian = gs.einsum("...ij,...j->...ij", rotated_sylvester, eigvals)
        rotated_hessian = Matrices.mul(rotated_hessian, rotated_sylvester)
        hessian = Matrices.mul(eigvecs, rotated_hessian, transconj_eigvecs)

        return base_point + tangent_vec + hessian

    def log(self, point, base_point):
        """Compute the Bures-Wasserstein logarithm map.

        Compute the Riemannian logarithm at point base_point,
        of point wrt the Bures-Wasserstein metric.
        This gives a tangent vector at point base_point.

        Parameters
        ----------
        point : array-like, shape=[..., n, n]
            Point.
        base_point : array-like, shape=[..., n, n]
            Base point.

        Returns
        -------
        log : array-like, shape=[..., n, n]
            Riemannian logarithm.
        """
        sqrt_bp, inv_sqrt_bp = powermh(base_point, [0.5, -0.5])
        pdt = powermh(Matrices.mul(sqrt_bp, point, sqrt_bp), 0.5)
        sqrt_product = Matrices.mul(sqrt_bp, pdt, inv_sqrt_bp)
        transconj_sqrt_product = ComplexMatrices.transconjugate(sqrt_product)
        return sqrt_product + transconj_sqrt_product - 2 * base_point

    def squared_dist(self, point_a, point_b):
        """Compute the Bures-Wasserstein squared distance.

        Compute the Riemannian squared distance between point_a and point_b.

        Parameters
        ----------
        point_a : array-like, shape=[..., n, n]
            Point.
        point_b : array-like, shape=[..., n, n]
            Point.

        Returns
        -------
        squared_dist : array-like, shape=[...]
            Riemannian squared distance.
        """
        product = gs.matmul(point_a, point_b)
        sqrt_product = gs.linalg.sqrtm(product)
        trace_a = gs.trace(point_a)
        trace_b = gs.trace(point_b)
        trace_prod = gs.trace(sqrt_product)

        squared_dist = gs.real(trace_a + trace_b - 2.0 * trace_prod)

        return gs.where(squared_dist < 0.0, 0.0, squared_dist)

    def parallel_transport(
        self,
        tangent_vec,
        base_point,
        direction=None,
        end_point=None,
        n_steps=10,
        step="rk4",
    ):
        r"""Compute the parallel transport of a tangent vec along a geodesic.

        Approximation of the solution of the parallel transport of a tangent
        vector a along the geodesic defined by :math:`t \mapsto exp_{(
        base\_point)}(t* tangent\_vec\_b)`. The parallel transport equation is
        formulated in this case in [TP2021]_.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., n, n]
            Tangent vector at `base_point` to transport.
        base_point : array-like, shape=[..., n, n]
            Initial point of the geodesic.
        direction : array-like, shape=[..., n, n]
            Tangent vector at `base_point`, initial velocity of the geodesic to
            transport along.
        end_point : array-like, shape=[..., n, n]
            Point to transport to.
            Optional, default: None.
        n_steps : int
            Number of steps to use to approximate the solution of the
            ordinary differential equation.
            Optional, default: 100
        step : str, {'euler', 'rk2', 'rk4'}
            Scheme to use in the integration scheme.
            Optional, default: 'rk4'.

        Returns
        -------
        transported :  array-like, shape=[..., n, n]
            Transported tangent vector at `exp_(base_point)(tangent_vec_b)`.

        See Also
        --------
        Integration module: geomstats.integrator
        """
        if end_point is None:
            end_point = self.exp(direction, base_point)

        horizontal_lift_a = gs.linalg.solve_sylvester(
            base_point, base_point, tangent_vec
        )

        square_root_bp, inverse_square_root_bp = powermh(base_point, [0.5, -0.5])
        end_point_lift = Matrices.mul(square_root_bp, end_point, square_root_bp)
        square_root_lift = powermh(end_point_lift, 0.5)

        horizontal_velocity = gs.matmul(inverse_square_root_bp, square_root_lift)
        partial_horizontal_velocity = Matrices.mul(horizontal_velocity, square_root_bp)
        partial_horizontal_velocity += ComplexMatrices.transconjugate(
            partial_horizontal_velocity
        )

        def force(state, time):
            horizontal_geodesic_t = (
                1 - time
            ) * square_root_bp + time * horizontal_velocity
            geodesic_t = (
                (1 - time) ** 2 * base_point
                + time * (1 - time) * partial_horizontal_velocity
                + time**2 * end_point
            )

            align = Matrices.mul(
                horizontal_geodesic_t,
                ComplexMatrices.transconjugate(horizontal_velocity - square_root_bp),
                state,
            )
            right = align + ComplexMatrices.transconjugate(align)
            return gs.linalg.solve_sylvester(geodesic_t, geodesic_t, -right)

        flow = integrate(force, horizontal_lift_a, n_steps=n_steps, step=step)
        final_align = Matrices.mul(end_point, flow[-1])
        return final_align + ComplexMatrices.transconjugate(final_align)

    def injectivity_radius(self, base_point):
        """Compute the upper bound of the injectivity domain.

        This is the smallest eigen value of the base point.

        Parameters
        ----------
        base_point : array-like, shape=[..., n, n]
            Point on the manifold.

        Returns
        -------
        radius : float
            Injectivity radius.
        """
        eigen_values = gs.linalg.eigvalsh(base_point)
        return eigen_values[..., 0] ** 0.5


class HPDEuclideanMetric(ComplexMatricesMetric):
    """Class for the Euclidean metric on the HPD manifold."""

    @staticmethod
    def exp_domain(tangent_vec, base_point):
        """Compute the domain of the Euclidean exponential map.

        Compute the real interval of time where the Euclidean geodesic starting
        at point `base_point` in direction `tangent_vec` is defined.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., n, n]
            Tangent vector at base point.
        base_point : array-like, shape=[..., n, n]
            Base point.

        Returns
        -------
        exp_domain : array-like, shape=[..., 2]
            Interval of time where the geodesic is defined.
        """
        invsqrt_base_point = powermh(base_point, -0.5)
        reduced_vec = gs.matmul(invsqrt_base_point, tangent_vec)
        reduced_vec = gs.matmul(reduced_vec, invsqrt_base_point)
        eigvals = gs.linalg.eigvalsh(reduced_vec)
        min_eig = gs.amin(eigvals, axis=-1)
        max_eig = gs.amax(eigvals, axis=-1)
        inf_value = gs.where(max_eig <= 0.0, gs.array(-math.inf), -1.0 / max_eig)
        sup_value = gs.where(min_eig >= 0.0, gs.array(-math.inf), -1.0 / min_eig)
        return gs.stack((inf_value, sup_value), axis=-1)

    def injectivity_radius(self, base_point):
        """Compute the upper bound of the injectivity domain.

        This is the smallest eigen value of the base point.

        Parameters
        ----------
        base_point : array-like, shape=[..., n, n]
            Point on the manifold.

        Returns
        -------
        radius : float
            Injectivity radius.
        """
        eigen_values = gs.linalg.eigvalsh(base_point)
        return eigen_values[..., 0]


class HPDLogEuclideanMetric(PullbackDiffeoMetric):
    """Class for the Log-Euclidean metric on the HPD manifold."""

    def __init__(self, space, image_space=None):
        if image_space is None:
            image_space = HermitianMatrices(n=space.n)
        diffeo = SymMatrixLog()
        super().__init__(space, diffeo, image_space)



"""Pytorch based computation backend."""

from collections.abc import Iterable as _Iterable

import numpy as _np
import torch as _torch
from torch import (
    arange,
    argmin,
    clip,
    complex64,
    complex128,
    conj,
    empty,
    empty_like,
    erf,
    eye,
    flatten,
    float32,
    float64,
    greater,
    hstack,
    int32,
    int64,
    isnan,
    kron,
    less,
    logical_or,
    mean,
    meshgrid,
    moveaxis,
    ones,
    ones_like,
    polygamma,
    quantile,
    reshape,
    scatter_add,
    searchsorted,
    stack,
    trapz,
    uint8,
    unique,
    vstack,
    zeros,
    zeros_like,
)
from torch import broadcast_tensors as broadcast_arrays
from torch import repeat_interleave as repeat
from torch.special import gammaln as _gammaln

from .._backend_config import pytorch_atol as atol
from .._backend_config import pytorch_rtol as rtol
from . import (
    autodiff,  # NOQA
    linalg,  # NOQA
    random,  # NOQA
)
from ._common import array, cast, from_numpy
from ._dtype import (
    _add_default_dtype_by_casting,
    _box_binary_scalar,
    _box_unary_scalar,
    _preserve_input_dtype,
    as_dtype,
    get_default_cdtype,
    get_default_dtype,
    is_bool,
    is_complex,
    is_floating,
    set_default_dtype,
)

_DTYPES = {
    int32: 0,
    int64: 1,
    float32: 2,
    float64: 3,
    complex64: 4,
    complex128: 5,
}


abs = _box_unary_scalar(target=_torch.abs)
angle = _box_unary_scalar(target=_torch.angle)
arccos = _box_unary_scalar(target=_torch.arccos)
arccosh = _box_unary_scalar(target=_torch.arccosh)
arcsin = _box_unary_scalar(target=_torch.arcsin)
arctanh = _box_unary_scalar(target=_torch.arctanh)
ceil = _box_unary_scalar(target=_torch.ceil)
cos = _box_unary_scalar(target=_torch.cos)
cosh = _box_unary_scalar(target=_torch.cosh)
exp = _box_unary_scalar(target=_torch.exp)
floor = _box_unary_scalar(target=_torch.floor)
log = _box_unary_scalar(target=_torch.log)
real = _box_unary_scalar(target=_torch.real)
sign = _box_unary_scalar(target=_torch.sign)
sin = _box_unary_scalar(target=_torch.sin)
sinh = _box_unary_scalar(target=_torch.sinh)
sqrt = _box_unary_scalar(target=_torch.sqrt)
tan = _box_unary_scalar(target=_torch.tan)
tanh = _box_unary_scalar(target=_torch.tanh)


arctan2 = _box_binary_scalar(target=_torch.atan2)
mod = _box_binary_scalar(target=_torch.remainder, box_x2=False)
power = _box_binary_scalar(target=_torch.pow, box_x2=False)


std = _preserve_input_dtype(_add_default_dtype_by_casting(target=_torch.std))


def has_autodiff():
    """If allows for automatic differentiation.

    Returns
    -------
    has_autodiff : bool
    """
    return True


def matmul(x, y, out=None):
    for array_ in [x, y]:
        if array_.ndim == 1:
            raise ValueError("ndims must be >=2")

    x, y = convert_to_wider_dtype([x, y])
    return _torch.matmul(x, y, out=out)


def to_numpy(x):
    return x.numpy()


def one_hot(labels, num_classes):
    if not _torch.is_tensor(labels):
        labels = _torch.LongTensor(labels)
    return _torch.nn.functional.one_hot(labels, num_classes).type(_torch.uint8)


def argmax(a, **kwargs):
    if a.dtype == _torch.bool:
        return _torch.as_tensor(_np.argmax(a.data.numpy(), **kwargs))
    return _torch.argmax(a, **kwargs)


def convert_to_wider_dtype(tensor_list):
    dtype_list = [_DTYPES.get(x.dtype, -1) for x in tensor_list]
    if len(set(dtype_list)) == 1:
        return tensor_list

    wider_dtype_index = max(dtype_list)

    wider_dtype = list(_DTYPES.keys())[wider_dtype_index]

    tensor_list = [cast(x, dtype=wider_dtype) for x in tensor_list]
    return tensor_list


def less_equal(x, y, **kwargs):
    if not _torch.is_tensor(x):
        x = _torch.tensor(x)
    if not _torch.is_tensor(y):
        y = _torch.tensor(y)
    return _torch.le(x, y, **kwargs)


def split(x, indices_or_sections, axis=0):
    if isinstance(indices_or_sections, int):
        indices_or_sections = x.shape[axis] // indices_or_sections
        return _torch.split(x, indices_or_sections, dim=axis)
    indices_or_sections = _np.array(indices_or_sections)
    intervals_length = indices_or_sections[1:] - indices_or_sections[:-1]
    last_interval_length = x.shape[axis] - indices_or_sections[-1]
    if last_interval_length > 0:
        intervals_length = _np.append(intervals_length, last_interval_length)
    intervals_length = _np.insert(intervals_length, 0, indices_or_sections[0])
    return _torch.split(x, tuple(intervals_length), dim=axis)


def logical_and(x, y):
    if _torch.is_tensor(x) or _torch.is_tensor(y):
        return x * y
    return x and y


def any(x, axis=None):
    if not _torch.is_tensor(x):
        x = _torch.tensor(x)
    if axis is None:
        return _torch.any(x)
    if isinstance(axis, int):
        return _torch.any(x.bool(), axis)
    if len(axis) == 1:
        return _torch.any(x, *axis)
    axis = list(axis)
    for i_axis, one_axis in enumerate(axis):
        if one_axis < 0:
            axis[i_axis] = ndim(x) + one_axis
    new_axis = tuple(k - 1 if k >= 0 else k for k in axis[1:])
    return any(_torch.any(x.bool(), axis[0]), new_axis)


def flip(x, axis):
    if isinstance(axis, int):
        axis = [axis]
    if axis is None:
        axis = list(range(x.ndim))
    return _torch.flip(x, dims=axis)


def concatenate(seq, axis=0, out=None):
    seq = convert_to_wider_dtype(seq)
    return _torch.cat(seq, dim=axis, out=out)


def all(x, axis=None):
    if not _torch.is_tensor(x):
        x = _torch.tensor(x)
    if axis is None:
        return x.bool().all()
    if isinstance(axis, int):
        return _torch.all(x.bool(), axis)
    if len(axis) == 1:
        return _torch.all(x, *axis)
    axis = list(axis)
    for i_axis, one_axis in enumerate(axis):
        if one_axis < 0:
            axis[i_axis] = ndim(x) + one_axis
    new_axis = tuple(k - 1 if k >= 0 else k for k in axis[1:])
    return all(_torch.all(x.bool(), axis[0]), new_axis)


def get_slice(x, indices):
    """Return a slice of an array, following Numpy's style.

    Parameters
    ----------
    x : array-like, shape=[dim]
        Initial array.
    indices : iterable(iterable(int))
        Indices which are kept along each axis, starting from 0.

    Returns
    -------
    slice : array-like
        Slice of x given by indices.

    Notes
    -----
    This follows Numpy's convention: indices are grouped by axis.

    Examples
    --------
    >>> a = torch.tensor(range(30)).reshape(3,10)
    >>> get_slice(a, ((0, 2), (8, 9)))
    tensor([8, 29])
    """
    return x[indices]


def allclose(a, b, atol=atol, rtol=rtol):
    if not isinstance(a, _torch.Tensor):
        a = _torch.tensor(a)
    if not isinstance(b, _torch.Tensor):
        b = _torch.tensor(b)
    a = to_ndarray(a.float(), to_ndim=1)
    b = to_ndarray(b.float(), to_ndim=1)
    n_a = a.shape[0]
    n_b = b.shape[0]
    nb_dim = a.dim()
    if n_a > n_b:
        reps = (int(n_a / n_b),) + (nb_dim - 1) * (1,)
        b = tile(b, reps)
    elif n_a < n_b:
        reps = (int(n_b / n_a),) + (nb_dim - 1) * (1,)
        a = tile(a, reps)
    return _torch.allclose(a, b, atol=atol, rtol=rtol)


def shape(val):
    if not is_array(val):
        val = array(val)
    return val.shape


def amax(a, axis=None):
    if axis is None:
        return _torch.max(array(a))
    return _torch.max(array(a), dim=axis).values


def maximum(a, b):
    return _torch.max(array(a), array(b))


def minimum(a, b):
    return _torch.min(array(a), array(b))


def to_ndarray(x, to_ndim, axis=0):
    if not _torch.is_tensor(x):
        x = array(x)

    if x.dim() == to_ndim - 1:
        x = _torch.unsqueeze(x, dim=axis)
    return x


def broadcast_to(x, shape):
    if not _torch.is_tensor(x):
        x = _torch.tensor(x)
    return x.expand(shape)


def isclose(x, y, rtol=rtol, atol=atol):
    if not _torch.is_tensor(x):
        x = _torch.tensor(x)
    if not _torch.is_tensor(y):
        y = _torch.tensor(y)
    return _torch.isclose(x, y, atol=atol, rtol=rtol)


def sum(x, axis=None, keepdims=None, dtype=None):
    if axis is None:
        if keepdims is None:
            return _torch.sum(x, dtype=dtype)
        return _torch.sum(x, keepdim=keepdims, dtype=dtype)
    if keepdims is None:
        return _torch.sum(x, dim=axis, dtype=dtype)
    return _torch.sum(x, dim=axis, keepdim=keepdims, dtype=dtype)


def einsum(equation, *inputs):
    input_tensors_list = [arg if is_array(arg) else array(arg) for arg in inputs]
    input_tensors_list = convert_to_wider_dtype(input_tensors_list)

    return _torch.einsum(equation, *input_tensors_list)


def transpose(x, axes=None):
    if axes:
        return x.permute(axes)
    if x.dim() == 1:
        return x
    if x.dim() > 2 and axes is None:
        return x.permute(tuple(range(x.ndim)[::-1]))
    return x.t()


def squeeze(x, axis=None):
    if not is_array(x):
        return x
    if axis is None:
        return _torch.squeeze(x)
    return _torch.squeeze(x, dim=axis)


def trace(x):
    if x.ndim == 2:
        return _torch.trace(x)

    return _torch.einsum("...ii", x)


def linspace(start, stop, num=50, endpoint=True, dtype=None):
    start_is_array = _torch.is_tensor(start)
    stop_is_array = _torch.is_tensor(stop)

    if not (start_is_array or stop_is_array) and endpoint:
        return _torch.linspace(start=start, end=stop, steps=num, dtype=dtype)

    if not start_is_array:
        start = _torch.tensor(start)
    if not stop_is_array:
        stop = _torch.tensor(stop)
    start, stop = _torch.broadcast_tensors(start, stop)
    result_shape = (num, *start.shape)
    start = _torch.flatten(start)
    stop = _torch.flatten(stop)

    if endpoint:
        result = _torch.vstack(
            [
                _torch.linspace(start=start[i], end=stop[i], steps=num, dtype=dtype)
                for i in range(start.shape[0])
            ]
        ).T
    else:
        result = _torch.vstack(
            [
                _torch.arange(
                    start=start[i],
                    end=stop[i],
                    step=(stop[i] - start[i]) / num,
                    dtype=dtype,
                )
                for i in range(start.shape[0])
            ]
        ).T

    return _torch.reshape(result, result_shape)


def equal(a, b, **kwargs):
    if not is_array(a):
        a = array(a)

    if not is_array(b):
        b = array(b)
    return _torch.eq(a, b, **kwargs)


def diag_indices(*args, **kwargs):
    return tuple(map(_torch.from_numpy, _np.diag_indices(*args, **kwargs)))


def tril(mat, k=0):
    return _torch.tril(mat, diagonal=k)


def triu(mat, k=0):
    return _torch.triu(mat, diagonal=k)


def tril_indices(n, k=0, m=None):
    if m is None:
        m = n
    return _torch.tril_indices(row=n, col=m, offset=k)


def triu_indices(n, k=0, m=None):
    if m is None:
        m = n
    return _torch.triu_indices(row=n, col=m, offset=k)


def tile(x, y):
    if not _torch.is_tensor(x):
        x = _torch.tensor(x)
    return x.repeat(y)


def expand_dims(x, axis=0):
    return _torch.unsqueeze(x, dim=axis)


def ndim(x):
    return x.dim()


def hsplit(x, indices_or_section):
    if isinstance(indices_or_section, int):
        indices_or_section = x.shape[-1] // indices_or_section
    return _torch.split(x, indices_or_section, dim=-1)


def diagonal(x, offset=0, axis1=0, axis2=1):
    return _torch.diagonal(x, offset=offset, dim1=axis1, dim2=axis2)


def set_diag(x, new_diag):
    """Set the diagonal along the last two axis.

    Parameters
    ----------
    x : array-like, shape=[dim]
        Initial array.
    new_diag : array-like, shape=[dim[-2]]
        Values to set on the diagonal.

    Returns
    -------
    None

    Notes
    -----
    This mimics tensorflow.linalg.set_diag(x, new_diag), when new_diag is a
    1-D array, but modifies x instead of creating a copy.
    """
    arr_shape = x.shape
    off_diag = (1 - _torch.eye(arr_shape[-1])) * x
    diag = _torch.einsum("ij,...i->...ij", _torch.eye(new_diag.shape[-1]), new_diag)
    return diag + off_diag


def prod(x, axis=None):
    if axis is None:
        axis = 0
    return _torch.prod(x, axis)


def where(condition, x=None, y=None):
    if not _torch.is_tensor(condition):
        condition = array(condition)

    if x is None and y is None:
        return _torch.where(condition)
    if not _torch.is_tensor(x):
        x = _torch.tensor(x)
    if not _torch.is_tensor(y):
        y = _torch.tensor(y)
    y = cast(y, x.dtype)
    return _torch.where(condition, x, y)


def _is_boolean(x):
    if isinstance(x, bool):
        return True
    if isinstance(x, (tuple, list)):
        return _is_boolean(x[0])
    if _torch.is_tensor(x):
        return x.dtype in [_torch.bool, _torch.uint8]
    return False


def _is_iterable(x):
    if isinstance(x, (list, tuple)):
        return True
    if _torch.is_tensor(x):
        return ndim(x) > 0
    return False


def assignment(x, values, indices, axis=0):
    """Assign values at given indices of an array.

    Parameters
    ----------
    x: array-like, shape=[dim]
        Initial array.
    values: {float, list(float)}
        Value or list of values to be assigned.
    indices: {int, tuple, list(int), list(tuple)}
        Single int or tuple, or list of ints or tuples of indices where value
        is assigned.
        If the length of the tuples is shorter than ndim(x), values are
        assigned to each copy along axis.
    axis: int, optional
        Axis along which values are assigned, if vectorized.

    Returns
    -------
    x_new : array-like, shape=[dim]
        Copy of x with the values assigned at the given indices.

    Notes
    -----
    If a single value is provided, it is assigned at all the indices.
    If a list is given, it must have the same length as indices.
    """
    x_new = copy(x)

    use_vectorization = hasattr(indices, "__len__") and len(indices) < ndim(x)
    if _is_boolean(indices):
        x_new[indices] = values
        return x_new
    zip_indices = _is_iterable(indices) and _is_iterable(indices[0])
    len_indices = len(indices) if _is_iterable(indices) else 1
    if zip_indices:
        indices = tuple(zip(*indices))
    if not use_vectorization:
        if not zip_indices:
            len_indices = len(indices) if _is_iterable(indices) else 1
        len_values = len(values) if _is_iterable(values) else 1
        if len_values > 1 and len_values != len_indices:
            raise ValueError("Either one value or as many values as indices")
        x_new[indices] = values
    else:
        indices = tuple(list(indices[:axis]) + [slice(None)] + list(indices[axis:]))
        x_new[indices] = values
    return x_new


def assignment_by_sum(x, values, indices, axis=0):
    """Add values at given indices of an array.

    Parameters
    ----------
    x: array-like, shape=[dim]
        Initial array.
    values: {float, list(float)}
        Value or list of values to be assigned.
    indices: {int, tuple, list(int), list(tuple)}
        Single int or tuple, or list of ints or tuples of indices where value
        is assigned.
        If the length of the tuples is shorter than ndim(x), values are
        assigned to each copy along axis.
    axis: int, optional
        Axis along which values are assigned, if vectorized.

    Returns
    -------
    x_new : array-like, shape=[dim]
        Copy of x with the values assigned at the given indices.

    Notes
    -----
    If a single value is provided, it is assigned at all the indices.
    If a list is given, it must have the same length as indices.
    """
    x_new = copy(x)
    values = array(values)
    use_vectorization = hasattr(indices, "__len__") and len(indices) < ndim(x)
    if _is_boolean(indices):
        x_new[indices] += values
        return x_new
    zip_indices = _is_iterable(indices) and _is_iterable(indices[0])
    if zip_indices:
        indices = list(zip(*indices))
    if not use_vectorization:
        len_indices = len(indices) if _is_iterable(indices) else 1
        len_values = len(values) if _is_iterable(values) else 1
        if len_values > 1 and len_values != len_indices:
            raise ValueError("Either one value or as many values as indices")
        x_new[indices] += values
    else:
        indices = tuple(list(indices[:axis]) + [slice(None)] + list(indices[axis:]))
        x_new[indices] += values
    return x_new


def copy(x):
    return x.clone()


def cumsum(x, axis=None, dtype=None):
    if not _torch.is_tensor(x):
        x = array(x, dtype=dtype)
    if axis is None:
        return x.flatten().cumsum(dim=0, dtype=dtype)
    return _torch.cumsum(x, dim=axis, dtype=dtype)


def cumprod(x, axis=None, dtype=None):
    if axis is None:
        axis = 0
    return _torch.cumprod(x, axis, dtype=dtype)


def array_from_sparse(indices, data, target_shape):
    """Create an array of given shape, with values at specific indices.

    The rest of the array will be filled with zeros.

    Parameters
    ----------
    indices : iterable(tuple(int))
        Index of each element which will be assigned a specific value.
    data : iterable(scalar)
        Value associated at each index.
    target_shape : tuple(int)
        Shape of the output array.

    Returns
    -------
    a : array, shape=target_shape
        Array of zeros with specified values assigned to specified indices.
    """
    return _torch.sparse.FloatTensor(
        _torch.LongTensor(indices).t(),
        array(data),
        _torch.Size(target_shape),
    ).to_dense()


def vectorize(x, pyfunc, multiple_args=False, **kwargs):
    if multiple_args:
        return stack(list(map(lambda y: pyfunc(*y), zip(*x))))
    return stack(list(map(pyfunc, x)))


def vec_to_diag(vec):
    return _torch.diag_embed(vec, offset=0)


def tril_to_vec(x, k=0):
    n = x.shape[-1]
    rows, cols = tril_indices(n, k=k)
    return x[..., rows, cols]


def triu_to_vec(x, k=0):
    n = x.shape[-1]
    rows, cols = triu_indices(n, k=k)
    return x[..., rows, cols]


def mat_from_diag_triu_tril(diag, tri_upp, tri_low):
    """Build matrix from given components.

    Forms a matrix from diagonal, strictly upper triangular and
    strictly lower traingular parts.

    Parameters
    ----------
    diag : array_like, shape=[..., n]
    tri_upp : array_like, shape=[..., (n * (n - 1)) / 2]
    tri_low : array_like, shape=[..., (n * (n - 1)) / 2]

    Returns
    -------
    mat : array_like, shape=[..., n, n]
    """
    diag, tri_upp, tri_low = convert_to_wider_dtype([diag, tri_upp, tri_low])

    n = diag.shape[-1]
    (i,) = diag_indices(n, ndim=1)
    j, k = triu_indices(n, k=1)
    mat = _torch.zeros((diag.shape + (n,)), dtype=diag.dtype)
    mat[..., i, i] = diag
    mat[..., j, k] = tri_upp
    mat[..., k, j] = tri_low
    return mat


def divide(a, b, ignore_div_zero=False):
    if ignore_div_zero is False:
        return _torch.divide(a, b)
    quo = _torch.divide(a, b)
    return _torch.nan_to_num(quo, nan=0.0, posinf=0.0, neginf=0.0)


def ravel_tril_indices(n, k=0, m=None):
    if m is None:
        size = (n, n)
    else:
        size = (n, m)
    idxs = _np.tril_indices(n, k, m)
    return _torch.from_numpy(_np.ravel_multi_index(idxs, size))


def sort(a, axis=-1):
    sorted_a, _ = _torch.sort(a, dim=axis)
    return sorted_a


def amin(a, axis=-1):
    (values, _) = _torch.min(a, dim=axis)
    return values


def take(a, indices, axis=0):
    if not _torch.is_tensor(indices):
        indices = _torch.as_tensor(indices)

    return _torch.squeeze(_torch.index_select(a, axis, indices))


def _unnest_iterable(ls):
    out = []
    if isinstance(ls, _Iterable):
        for inner_ls in ls:
            out.extend(_unnest_iterable(inner_ls))
    else:
        out.append(ls)

    return out


def pad(a, pad_width, mode="constant", constant_values=0.0):
    return _torch.nn.functional.pad(
        a, _unnest_iterable(reversed(pad_width)), mode=mode, value=constant_values
    )


def is_array(x):
    return _torch.is_tensor(x)


def outer(a, b):
    # TODO: improve for torch > 1.9 (dims=0 fails in 1.9)
    return _torch.einsum("...i,...j->...ij", a, b)


def matvec(A, b):
    A, b = convert_to_wider_dtype([A, b])

    if A.ndim == 2 and b.ndim == 1:
        return _torch.mv(A, b)

    if b.ndim == 1:  # A.ndim > 2
        return _torch.matmul(A, b)

    if A.ndim == 2:  # b.ndim > 1
        return _torch.matmul(A, b.T).T

    return _torch.einsum("...ij,...j->...i", A, b)


def dot(a, b):
    a, b = convert_to_wider_dtype([a, b])

    if a.ndim == 1 and b.ndim == 1:
        return _torch.dot(a, b)

    if b.ndim == 1:
        return _torch.tensordot(a, b, dims=1)

    if a.ndim == 1:
        return _torch.tensordot(a, b.T, dims=1)

    return _torch.einsum("...i,...i->...", a, b)


def cross(a, b):
    if a.shape != b.shape:
        a, b = broadcast_arrays(a, b)
    return _torch.cross(*convert_to_wider_dtype([a, b]), dim=-1)


def gamma(a):
    return _torch.exp(_gammaln(a))


def imag(a):
    if not _torch.is_tensor(a):
        a = _torch.tensor(a)
    if is_complex(a):
        return _torch.imag(a)
    return _torch.zeros(a.shape, dtype=a.dtype)



import pytest

import geomstats.backend as gs
from geomstats.test_cases.learning._base import BaseEstimatorTestCase


class TangentPCATestCase(BaseEstimatorTestCase):
    @pytest.mark.random
    def test_fit_inverse_transform(self, n_samples, atol):
        X = expected = self.data_generator.random_point(n_samples)

        tangent_projected_data = self.estimator.fit_transform(X)
        res = self.estimator.inverse_transform(tangent_projected_data)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.random
    def test_fit_transform_and_transform_after_fit(self, n_samples, atol):
        X = self.data_generator.random_point(n_samples)

        res = self.estimator.fit_transform(X)
        res_ = self.estimator.fit(X).transform(X)
        self.assertAllClose(res, res_, atol=atol)

    @pytest.mark.random
    def test_n_components(self, n_samples):
        X = self.data_generator.random_point(n_samples)

        n_components_0 = self.estimator.n_components

        n_components = 2
        self.estimator.n_components = n_components

        self.estimator.fit(X)
        self.assertEqual(self.estimator.n_components_, n_components)

        self.estimator.n_components = n_components_0

    @pytest.mark.random
    def test_n_components_explained_variance_ratio(self, n_samples, atol):
        X = self.data_generator.random_point(n_samples)

        n_components_0 = self.estimator.n_components

        target = 0.9
        self.estimator.n_components = target

        self.estimator.fit(X)
        res = gs.cumsum(self.estimator.explained_variance_ratio_)[-1]
        self.assertTrue(res > target - atol)

        self.estimator.n_components = n_components_0

    @pytest.mark.random
    def test_n_components_mle(self, n_samples):
        if self.estimator.space.point_ndim > 1:
            return

        X = self.data_generator.random_point(n_samples)

        n_components_0 = self.estimator.n_components

        n_components = "mle"
        self.estimator.n_components = n_components

        self.estimator.fit(X)
        self.assertEqual(self.estimator.n_features_, gs.shape(X)[1])

        self.estimator.n_components = n_components_0



from geomstats.test_cases.geometry.riemannian_metric import RiemannianMetricTestCase


class SasakiMetricTestCase(RiemannianMetricTestCase):
    def test_geodesic_discrete(self, initial_point, end_point, expected, atol):
        res = self.space.metric.geodesic_discrete(initial_point, end_point)
        self.assertAllClose(res, expected, atol=atol)
